# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import nose.tools as nt

import sympy

from fastr.core.dimension import HasDimensions, Dimension
from fastr import exceptions


class Dimensional(HasDimensions):
    """
    Test implementation of a HasDimensions class for testing.
    """
    SIZE = (42, 1337, sympy.Symbol('X'))
    NAME = ('answer', 'l33t', 'vague')

    def __init__(self):
        self._dimensions = tuple(Dimension(x, y) for x, y in zip(self.NAME, self.SIZE))

    @property
    def dimensions(self):
        return self._dimensions


class TestDimension():
    """
    Test for the Dimension
    """
    def setup(self):
        self.cleese = Dimension('john', 196)
        self.palin = Dimension('michael', 178)

    def test_name(self):
        nt.eq_(self.cleese.name, 'john')
        nt.eq_(self.palin.name, 'michael')

    def test_size(self):
        nt.eq_(self.cleese.size, 196)
        nt.eq_(self.palin.size, 178)

    @nt.raises(exceptions.FastrTypeError)
    def test_wrong_size_type_str(self):
        Dimension('eric', 'idle')

    @nt.raises(exceptions.FastrTypeError)
    def test_wrong_size_type_float(self):
        Dimension('graham', 1.88)

    @nt.raises(exceptions.FastrTypeError)
    def test_wrong_name_type_unicode(self):
        Dimension(u'terry', 173)

    @nt.raises(exceptions.FastrValueError)
    def test_wrong_value_size(self):
        Dimension('negative', -1)


class TestHasDimension():
    """
    Tests for the Dimenions and HasDimensions mixin
    """
    def setup(self):
        self.test_object = Dimensional()

    def test_size(self):
        nt.eq_(self.test_object.size, self.test_object.SIZE)

    def test_dimnames(self):
        nt.eq_(self.test_object.dimnames, self.test_object.NAME)



#!/usr/bin/python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The executionscript is the script that wraps around a tool executable. It
takes a job, builds the command, executes the command (while profiling it)
and collects the results.
"""

import datetime
import gzip
import os
try:
    import cPickle as pickle
except:
    import pickle
import sys
import traceback
import urlparse

FASTR_LOG_TYPE = 'console'
import fastr
from fastr import exceptions
from fastr.execution.job import Job, SinkJob, JobState
from fastr.utils.sysinfo import get_hostinfo, get_sysinfo

_MONITOR_INTERVAL = 1.0

# We will run untrusted subprocesses and want to log everything also in case
# of an error. Therefor there are a few bare-except blocks which are intended
# to ensure as much as possible logging in case of errors.
# pyline: disable=bare-except

fastr.toollist.populate()


def execute_job(job):
    """
    Execute a Job and save the result to disk

    :param job: the job to execute
    """
    fastr.log.info('FASTR loaded from {}'.format(fastr.__file__))
    job_dir = fastr.vfs.url_to_path(job.tmpdir)
    fastr.log.info('Set current directory to job output dir {}'.format(job_dir))
    os.chdir(job_dir)

    logfile_path = job.logfile
    fastr.log.info('Job log path: {}'.format(logfile_path))

    fastr.log.info('Running job {}\n  command: {} v{}\n  arguments: {}\n  outputs: {}\n'.format(job.jobid, job.tool_name, job.tool_version, job.input_arguments, job.output_arguments))
    try:
        job.status = JobState.running
        job.info_store['hostinfo'] = get_hostinfo()
        job.info_store['sysinfo_start'] = get_sysinfo()

        fastr.log.info('DRMAA info: {}'.format(job.info_store['hostinfo']['drmaa']))

        job.execute()

        # Document system information after
        fastr.log.info('Job subprocess finished')
        job.info_store['sysinfo_end'] = get_sysinfo()

        fastr.log.info('Start hashing results')
        start = datetime.datetime.now()
        job.hash_results()
        end = datetime.datetime.now()
        fastr.log.info('Finished hashing results in {} seconds'.format((end - start).total_seconds()))

    except:
        # Log errors to the info store
        if 'process' not in job.info_store:
            job.info_store['errors'].append(exceptions.FastrSubprocessNotFinished('There is no information that the subprocess finished properly: appears the job crashed before the subprocess registered as finished.').excerpt())
        elif job.info_store['process']['stderr'] != '':
            job.info_store['errors'].append(exceptions.FastrErrorInSubprocess(job.info_store['process']['stderr']).excerpt())
        exc_type, exception, trace = sys.exc_info()
        exc_info = traceback.format_exc()
        if isinstance(exception, exceptions.FastrError):
            job.info_store['errors'].append((exc_type.__name__, exception, exception.filename, exception.linenumber))
        else:
            trace = traceback.extract_tb(trace, 1)[0]
            job.info_store['errors'].append((exc_type.__name__, exception, trace[0], trace[1]))
        fastr.log.critical('Execution script encountered errors: {}'.format(exc_info))
        job.status = JobState.execution_failed
    else:
        job.status = JobState.execution_done
    finally:
        fastr.log.info('Writing job result to: {}'.format(logfile_path))
        with gzip.open(logfile_path, 'wb') as result_file:
            result_file.write(pickle.dumps(job))

        # If Sink Job write prov next to sink output.
        if isinstance(job, SinkJob):
            fastr.log.info('SinkJob input arguments: {}'.format(job.input_arguments))
            arguments = {key: value for key, value in job.input_arguments.items()}
            for output_value in arguments['output']:
                # TODO: Since we want to move to other means of storing prov documents, this is a quick and dirty fix for supporting url's with query information.
                url = urlparse.urlparse(job.substitute(output_value, datatype=type(arguments['input'].data.sequence_part()[0])))
                result_url = urlparse.urlunparse(url._replace(path=url.path+".prov.pickle.gz"))
                fastr.ioplugins.put_url(logfile_path, result_url)
        # Cleanup all plugins
        fastr.ioplugins.cleanup()


def main(joblist=None):
    """
    This is the main code. Wrapped inside a function to avoid the variables
    being seen as globals and to shut up pylint. Also if the joblist argument
    is given it can run any given job, otherwise it takes the first command
    line argument.
    """
    fastr.log.info('----- Execution script -----\n')
    if joblist is None:
        joblist = sys.argv[1]

    if os.path.exists(joblist) and os.path.isfile(joblist):
        fastr.log.info('Loading pickled command from file')
        start = datetime.datetime.now()
        with gzip.open(joblist, 'rb') as fin:
            joblist = pickle.load(fin)
        end = datetime.datetime.now()
        fastr.log.info('Finished loading pickle in {} seconds'.format((end - start).total_seconds()))

    # Both the error we raise and not able to iterate joblist will result in a
    # TypeError, so we catch both and give our own TypeError message
    try:
        if not isinstance(joblist, Job) and not all(isinstance(x, Job) for x in joblist):
            raise TypeError('Wrong type!')
    except TypeError:
        message = 'Argument should be a Job or a iterable of Jobs! (Found: {})'.format(joblist)
        fastr.log.critical(message)
        raise exceptions.FastrTypeError(message)

    fastr.log.info('Received command: {command}\n'.format(command=joblist))

    if isinstance(joblist, list) or isinstance(joblist, tuple):
        for job in joblist:
            execute_job(job)
    elif isinstance(joblist, Job):
        execute_job(joblist)
    fastr.log.info('---------------------------\n')


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import fastr
from fastr.core.version import Version
from fastr.data import url
from fastr.datatypes import URLType


class Directory(URLType):
    """
    DataType representing a directory.
    """
    description = 'A directory on the disk'
    extension = None

    def _validate(self):
        """
        Validate the value of the DataType.

        :return: flag indicating validity of Boolean
        :rtype: bool
        """
        value = self.value

        if url.isurl(self.value):
            value = url.get_path_from_url(value)

        try:
            return os.path.isdir(value)
        except ValueError:
            return False

    def __eq__(self, other):
        """
        Directories are equal by default as long as the validatity matches.

        :param Directory other: other to compare against
        :return: equality flag
        """
        return self.valid == other.valid

    def action(self, name):
        """
        This method makes sure the Directory exists. A Tool can indicate an
        action that should be called for an Output which will be called before
        execution.

        :param str name: name of the action to execute
        :return: None
        """
        if name is None:
            pass
        elif name == "ensure":
            if url.isurl(self.value):
                dir_name = url.get_path_from_url(self.value)
            else:
                dir_name = self.value

            fastr.log.debug('ensuring {} exists.'.format(dir_name))
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)
        else:
            fastr.log.warning("unknown action")



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys
import traceback
import fastr
import fastr.resources
from fastr.execution.executionpluginmanager import ExecutionPlugin, JobAction
from fastr.execution.job import Job, JobState


def run_job(job, job_status):
    try:
        fastr.log.debug('Running job {}'.format(job.jobid))
        job_status[job.jobid] = JobState.running

        command = [sys.executable,
                   os.path.join(fastr.config.executionscript),
                   job.commandfile]

        with open(job.stdoutfile, 'w') as fh_stdout, open(job.stderrfile, 'w') as fh_stderr:
            proc = subprocess.Popen(command, stdout=fh_stdout, stderr=fh_stderr)
            proc.wait()
            fastr.log.debug('Subprocess finished')
        fastr.log.debug('Finished {}'.format(job.jobid))
    except Exception:
        exc_type, _, trace = sys.exc_info()
        exc_info = traceback.format_exc()
        trace = traceback.extract_tb(trace, 1)[0]
        fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
        job.info_store['errors'].append((exc_type.__name__, exc_info, trace[0], trace[1]))
        job.status = JobState.failed

    return job


class BlockingExecution(ExecutionPlugin):
    """
    The blocking execution plugin is a special plugin which is meant for debug
    purposes. It will not queue jobs but immediately execute them inline,
    effectively blocking fastr until the Job is finished. It is the simplest
    execution plugin and can be used as a template for new plugins or for
    testing purposes.
    """
    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        super(BlockingExecution, self).__init__(finished_callback, cancelled_callback, status_callback)

    def cleanup(self):
        super(BlockingExecution, self).cleanup()

    def _job_finished(self, result):
        pass

    def _cancel_job(self, job):
        pass

    def _release_job(self, job):
        if not isinstance(job, Job):
            job = self.job_dict[job]
        self.queue_job(job)

    def _queue_job(self, job):
        # Check if the job is ready to run or must be held
        action = self.check_job_requirements(job.jobid)
        if action == JobAction.hold:
            fastr.log.debug('Holding {} until dependencies are met'.format(job.jobid))
            self.job_status[job.jobid] = job.status = JobState.hold
        elif action == JobAction.cancel:
            self.cancel_job(job)
        else:
            fastr.log.debug('Queueing {}'.format(job.jobid))
            run_job(job, self.job_status)
            self.job_finished(job, blocking=True)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os
import fastr


def function_wrapper(filepath, fnc_name, *args, **kwargs):
    fastr.log.debug('Starting function wrapper: {} {}'.format(filepath, fnc_name))
    # Prepend filename so importing will be possible
    filebase, extension = os.path.splitext(os.path.basename(filepath))
    fastr.log.debug('Loading module with {} {}'.format(filebase, filepath))

    if extension == '.py':
        temp_module = imp.load_source(filebase, filepath)
    elif extension == '.pyc':
        temp_module = imp.load_compiled(filebase, filepath)
    else:
        message = 'File does not have correct extension, must be .py or .pyc (is {})!'.format(extension)
        fastr.log.critical(message)
        raise ValueError(message)

    if hasattr(temp_module, fnc_name):
        fnc = getattr(temp_module, fnc_name)
    else:
        fastr.log.critical('Cannot get pickled function!')
        raise ValueError('Function not found in file!')

    fastr.log.debug('Calling function')
    return fnc(*args, **kwargs)


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from Queue import Queue, Empty
import subprocess
import sys
from threading import Thread
import traceback

import fastr
from fastr.execution.executionpluginmanager import ExecutionPlugin, JobAction
from fastr.execution.job import Job, JobState


class LinearExecution(ExecutionPlugin):
    """
    An execution engine that has a background thread that executes the jobs in
    order. The queue is a simple FIFO queue and there is one worker thread that
    operates in the background. This plugin is meant as a fallback when other
    plugins do not function properly. It does not multi-processing so it is
    safe to use in environments that do no support that.
    """
    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        super(LinearExecution, self).__init__(finished_callback, cancelled_callback, status_callback)
        self.job_queue = Queue()
        self.timeout = 1.0
        self._exec_thread = Thread(None, self.exec_worker, 'ExecWorker')
        self.running = True
        self._exec_thread.start()

    def _queue_job(self, job):
        # Check if the job is ready to run or must be held
        action = self.check_job_requirements(job.jobid)
        if action == JobAction.hold:
            fastr.log.debug('Holding {} until dependencies are met'.format(job.jobid))
            self.job_status[job.jobid] = job.status = JobState.hold
        elif action == JobAction.cancel:
            self.cancel_job(job)
        else:
            fastr.log.debug('Queueing {}'.format(job.jobid))
            self.job_queue.put(job)
            job.status = JobState.queued
            self.job_finished(job, blocking=True)

    def _release_job(self, job):
        if not isinstance(job, Job):
            job = self.job_dict[job]
        self.queue_job(job)

    def _cancel_job(self, job):
        fastr.log.debug('Cannot remove jobs from a LinearExecution queue!')

    def _job_finished(self, result):
        pass

    def cleanup(self):
        super(LinearExecution, self).cleanup()
        self.running = False
        self._exec_thread.join()
        self.job_status.clear()

    def exec_worker(self):
        while self.running:
            try:
                job = self.job_queue.get(True, self.timeout)
                job.status = JobState.running

                # Make sure subprocess causes no trouble
                try:
                    fastr.log.debug('Running job {}'.format(job.jobid))
                    job.status = JobState.running

                    command = [sys.executable,
                               os.path.join(fastr.config.executionscript),
                               job.commandfile]

                    with open(job.stdoutfile, 'w') as fh_stdout, open(job.stderrfile, 'w') as fh_stderr:
                        proc = subprocess.Popen(command, stdout=fh_stdout, stderr=fh_stderr)
                        proc.wait()
                        fastr.log.debug('Subprocess finished')
                    fastr.log.debug('Finished {}'.format(job.jobid))
                except Exception:
                    exc_type, _, trace = sys.exc_info()
                    exc_info = traceback.format_exc()
                    trace = traceback.extract_tb(trace, 1)[0]
                    fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
                    job.info_store['errors'].append((exc_type.__name__, exc_info, trace[0], trace[1]))
                    job.status = JobState.failed
                else:
                    job.status = JobState.finished
                finally:
                    self.job_finished(job)
                    self.job_queue.task_done()

            except Empty:
                pass



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import fastr
import gzip
import cPickle as pickle


def get_parser():
    parser = argparse.ArgumentParser(description="Get PROV information from the result pickle. When no options are given, the provenance syntax is printed to stdout in PROV-JSON format.")
    parser.add_argument('result', metavar='RESULTFILE', nargs='?', type=str, help='File of the job to execute (default ./__fastr_result__.pickle.gz)')
    parser.add_argument('-so', '--syntax-out-file', type=str, help="Write the syntax to file.")
    parser.add_argument('-sf', '--syntax-format', default='json', type=str, help="Choices are: [json], provn or xml")
    parser.add_argument('-i', '--indent', type=int, default=2, help="Indent size of the serialized documents.")
    parser.add_argument('-vo', '--visualize-out-file', type=str, help="Visualize the provenance. The most preferred format is svg. You can specify any format pydot supports. Specify the format by postfixing the filename with an extension.")
    return parser


def get_prov_document(result):
    with gzip.open(result, 'rb') as fh:
        data = pickle.load(fh)
    return data.provenance.document


def main(args=None, unknown_args=None):
    """
    Get PROV information from the result pickle.
    """
    if args is None and unknown_args is None:
        # No arguments were parsed yet, parse them now
        parser = get_parser()
        args, unknown_args = parser.parse_known_args()

    if args.result is not None:
        result = args.result
    else:
        curdir_contents = os.listdir('.')
        if '__fastr_result__.pickle.gz' in curdir_contents:
            result = '__fastr_result__.pickle.gz'
        else:
            fastr.log.critical('No result given and cannot find __fastr_result__.pickle.gz in current directory!')
            exit()

    # Get the prov document.
    provenance = get_prov_document(result)
    syntax_format = args.syntax_format

    if args.syntax_out_file:
        provenance.serialize(destination=args.syntax_out_file, format=syntax_format)
        fastr.log.info("Written the provenance syntax in {} format to: {}".format(syntax_format, args.syntax_out_file))

    if args.visualize_out_file:
        provenance.plot(args.visualize_out_file, show_element_attributes=False, show_relation_attributes=False)
        fastr.log.info("Vizualized the provenance in: {}".format(args.visualize_out_file))

    if not args.visualize_out_file and not args.syntax_out_file:
        # There is no output file give so let the serializer print to stdout.
        print(provenance.serialize(format=syntax_format, indent=2))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module to maintain a tool.

Exported classes:

* Tool -- A class encapsulating a tool.
* ParameterDescription -- The base class containing the shared description of
  a parameter (both input and ouput).
* InputParameterDescription -- A class containing the description of an input
  parameter.
* Output ParameterDescription -- A class containing the description of an
  output parameter.
"""

import itertools
import os
import platform
import re
from collections import namedtuple, OrderedDict

import jsonschema

import fastr
from fastr.core.baseplugin import PluginState
import fastr.core.target
import fastr.exceptions as exceptions
from fastr.core.serializable import Serializable
from fastr.core.version import Version
from fastr.utils.checksum import hashsum

try:
    from fastr.execution.environmentmodules import EnvironmentModules
    ENVIRONMENT_MODULES = EnvironmentModules(fastr.config.protected_modules)
    ENVIRONMENT_MODULES_LOADED = True
except exceptions.FastrValueError:
    ENVIRONMENT_MODULES_LOADED = False

from fastr.utils.dicteq import diffdict


class Tool(Serializable):
    """
    The class encapsulating a tool.
    """

    __dataschemafile__ = 'Tool.schema.json'

    test_spec = namedtuple('TestSpecification', ['input', 'command', 'output'])

    def __init__(self, doc=None):
        """
        Create a new Tool
        :param doc: path of toolfile or a dict containing the tool data
        :type doc: str or dict
        """

        # Cache value for target binary
        self._target = None

        if doc is None:
            return

        filename = None
        if not isinstance(doc, dict):
            fastr.log.debug('Trying to load file: {}'.format(doc))
            filename = os.path.expanduser(doc)
            filename = os.path.abspath(filename)
            doc = self._loadf(filename)

        # Check if the doc is a valid Tool structure
        try:
            # __unserializer__ is supplied by the serializable meta-class
            # pylint: disable=no-member
            doc = Tool.get_serializer().instantiate(doc)
        except jsonschema.ValidationError:
            fastr.log.error('Could not validate Tool data again the schema!')
            raise
        else:
            fastr.log.debug('Tool schema validated!')

        # Get attributes from root node
        self.filename = filename

        #: Identifier for the tool
        regex = r'^[A-Z][\w\d_]*$'
        if re.match(regex, doc['id']) is None:
            message = 'A tool id in Fastr should be UpperCamelCase as enforced' \
                      ' by regular expression {} (found {})'.format(regex, doc['id'])
            fastr.log.warning(message)

        self.id = doc['id']

        #: The namespace this tools lives in, this will be set by the ToolManager on load
        self.namespace = None

        #: Name of the tool, this should be a descriptive, human readable name.
        self.name = doc.get('name', self.id)

        #: Version of the tool, not of the underlying software
        self.version = Version(str(doc.get('version')))

        #: Class for of the Node to use
        self.node_class = doc.get('class', 'Node')

        if self.id is None or self.name is None or self.version is None:
            raise exceptions.FastrValueError('Tool should contain an id, name and version!')

        #: URL to website where this tool can be downloaded from
        self.url = doc.get('url', '')

        #: Description of the tool and it's functionality
        self.description = doc.get('description', '')

        #: List of authors of the tool. These people wrapped the executable but
        #: are not responsible for executable itself.
        self.authors = doc['authors']

        #: List of tags for this tool
        self.tags = doc.get('tags', [])

        # Parse references field and format them into a dictionary
        #: A list of documents and in depth reading about the methods used in this tool
        self.references = doc.get('references', [])

        #: Requirements for this Tool
        #:
        #: .. warning:: Not yet implemented
        self.requirements = None

        #: Test for this tool. A test should be a collection of inputs, parameters
        #: and outputs to verify the proper functioning of the Tool.
        #:
        #: The format of the tests is a list of namedtuples, that have 3 fields:
        #: - input: a dict of the input data
        #: - command: a list given the expected command-line arguments
        #: - output: a dict of the output data to validate
        #:
        #: .. warning:: Not yet implemented

        self.tests = [self.test_spec(**x) for x in doc['tests']]

        command = doc['command']

        # Find commands
        #: Command is a dictionary contain information about the command which is
        #: called by this Tool:
        #: command['interpreter'] holds the (possible) interpreter to use
        #: command['targets'] holds a per os/arch dictionary of files that should be executed
        #: command['url'] is the webpage of the command to be called
        #: command['version'] is the version of the command used
        #: command['description'] can help a description of the command
        #: command['authors'] lists the original authors of the command
        self.command = {
            'targets': command['targets'],
            'license': command.get('license', ''),
            'url': command.get('url', ''),
            'version': Version(str(command.get('version'))),
            'description': command.get('description', ''),
            'authors': command.get('authors', [])
        }

        if len(self.command['targets']) == 0:
            raise exceptions.FastrValueError("No targets defined in tool description.")

        #: This holds the citation you should use when publishing something based on this Tool
        self.cite = doc['cite']

        #: Man page for the Tool. Here usage and examples can be described in detail
        self.help = doc['help']

        #: Create the Interface based on the class specified in the tool file
        interface_class = fastr.interfaces[doc['interface'].get('class', 'FastrInterface')]

        if interface_class.status != PluginState.loaded:
            raise exceptions.FastrPluginNotLoaded('Required InterfacePlugin {} was not loaded properly (status {})'.format(interface_class.id,
                                                                                                                           interface_class.status))

        self.interface = interface_class(id_='{}_{}_interface'.format(self.id, self.command['version']), document=doc['interface'])

    @property
    def hash(self):
        return hashsum(self.__getstate__())

    @property
    def ns_id(self):
        """
        The namespace and id of the Tool
        """
        if self.namespace is None:
            return self.id
        else:
            return '{}.{}'.format(self.namespace, self.id)
    @property
    def fullid(self):
        """
        The full id of this tool
        """
        return 'fastr:///tools/{}/{}'.format(self.ns_id, self.version)

    @property
    def inputs(self):
        return self.interface.inputs

    @property
    def outputs(self):
        return self.interface.outputs

    @property
    def target(self):
        """
        The OS and arch matched target definition.
        """
        if self._target is not None:
            return self._target

        # Get platform and os
        arch = platform.architecture()[0].lower()
        os_ = platform.system().lower()

        matching_targets = [x for x in self.command['targets'] if x['os'] in [os_, '*'] and x['arch'] in [arch, '*']]

        if len(matching_targets) == 0:
            return None
        elif len(matching_targets) == 1:
            target = matching_targets[0]
        else:
            # This should give the optimal match
            for match in matching_targets:
                match['score'] = 0

                if match['os'] == os_:
                    match['score'] += 2

                if match['arch'] == arch:
                    match['score'] += 1

            matching_targets = sorted(matching_targets, reverse=True, key=lambda x: x['score'])
            fastr.log.debug('Sorted matches: {}'.format(matching_targets))
            target = matching_targets[0]

        cls = target.get('class', 'LocalBinaryTarget')
        if hasattr(fastr.core.target, cls):
            cls = getattr(fastr.core.target, cls)
        else:
            raise exceptions.FastrValueError('Cannot find target {}'.format(cls))

        # Create target from curdir
        old_curdir = os.path.abspath(os.curdir)
        os.chdir(self.path)

        # Check if the argument binary exists
        if 'binary' not in target:
            if 'bin' in target:
                target['binary'] = target['bin']
                del target['bin']
            else:
                fastr.exceptions.FastrValueError('Target does not contain required "binary" (or "bin") argument')

        self._target = cls(**target)
        os.chdir(old_curdir)

        return self._target

    def execute(self, payload=None, **kwargs):
        """
        Execute a Tool given the payload for a single run

        :param payload: the data to execute the Tool with
        :returns: The result of the execution
        :rtype: InterFaceResult
        """
        # Allow kwargs to be used instead of payload
        if payload is None:
            payload = {'inputs': {}, 'outputs': {}}
            for key, value in kwargs.items():
                if key in self.inputs and key in self.outputs:
                    raise exceptions.FastrValueError('Cannot figure out if "{}" is an input or output, please prefix with input_/output_ as needed')
                elif key in self.inputs:
                    payload['inputs'][key] = value
                elif key in self.outputs:
                    payload['outputs'][key] = value
                elif key.startswith('input_') and key[6:] in self.inputs:
                    payload['inputs'][key[6:]] = value
                elif key.startswith('output_') and key[7:] in self.outputs:
                    payload['outputs'][key[7:]] = value
                else:
                    raise exceptions.FastrValueError('Cannot match key "{}" to any input/output!'.format(key))

        # Make sure all values are wrapped in a tuple (for single values)
        for key, value in payload['inputs'].items():
            if not isinstance(value, (tuple, OrderedDict)):
                payload['inputs'][key] = (value,)
        for key, value in payload['outputs'].items():
            if not isinstance(value, (tuple, OrderedDict)):
                payload['outputs'][key] = (value,)

        target = self.target
        fastr.log.info('Target is {}'.format(target))

        if target is None:
            arch = platform.architecture()[0].lower()
            os_ = platform.system().lower()
            raise exceptions.FastrValueError('Cannot find a viable target for {}/{} on {} ({} bit)'.format(self.id, self.version, os_, arch))

        fastr.log.info('Using payload: {}'.format(payload))
        with target:
            result = self.interface.execute(target, payload)

        return result

    def __str__(self):
        """
        Get a string version for the Tool

        :return: the string version
        :rtype: str
        """
        return '<Tool: {} version: {}>'.format(self.id, self.version)

    def __repr__(self):
        """
        Get a string representation for the Tool. This will show the inputs
        and output defined in a table-like structure.

        :return: the string representation
        :rtype: str
        """
        if self.name is not None and len(self.name) > 0:
            name_part = ' ({})'.format(self.name)
        else:
            name_part = ''

        return_list = ["Tool {} v{}{}".format(self.id, str(self.command['version']), name_part)]

        # The "+ [8]" guarantees a minimum of 8 width and avoids empty lists
        width_input_keys = max([len(x.id) for x in self.inputs.values()] + [8])
        width_input_types = max([len(x.datatype) for x in self.inputs.values()] + [8]) + 2
        width_output_keys = max([len(x.id) for x in self.outputs.values()] + [8])
        width_output_types = max([len(x.datatype) for x in self.outputs.values()] + [8]) + 2

        return_list.append('{:^{}}  | {:^{}}'.format('Inputs', width_input_types + width_input_keys + 1,
                                                     'Outputs', width_output_types + width_output_keys + 1))
        return_list.append('-' * (width_input_keys + width_input_types + width_output_keys + width_output_types + 7))
        for input_, output in itertools.izip_longest(self.inputs.values(), self.outputs.values()):
            if input_ is None:
                input_id = ''
                input_type = ''
            else:
                input_id = input_.id
                input_type = '({})'.format(input_.datatype)

            if output is None:
                output_id = ''
                output_type = ''
            else:
                output_id = output.id
                output_type = '({})'.format(output.datatype)

            return_list.append('{:{}} {:{}}  |  {:{}} {:{}}'.format(input_id, width_input_keys,
                                                                    input_type, width_input_types,
                                                                    output_id, width_output_keys,
                                                                    output_type, width_output_types))

        return '\n'.join(return_list)

    def __eq__(self, other):
        """Compare two Tool instances with each other.

        :param other: the other instances to compare to
        :type other: Tool
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, Tool):
            return NotImplemented

        dict_self = dict(self.__dict__)
        del dict_self['_target']

        dict_other = dict(other.__dict__)
        del dict_other['_target']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the Tool

        :return: the state of the object
        :rtype dict:
        """
        state = {k: v for k, v in self.__dict__.items()}
        state['command'] = {k: v for k, v in self.command.items()}

        state['tests'] = [dict(vars(x)) for x in state['tests']]
        state['class'] = state['node_class']
        state['interface'] = self.interface.__getstate__()
        state['command']['version'] = str(self.command['version'])
        state['version'] = str(self.version)
        del state['node_class']
        del state['_target']

        return state

    def __setstate__(self, state):
        """
        Set the state of the Tool by the given state.

        :param dict state: The state to populate the object with
        """
        if 'filename' not in state:
            state['filename'] = None

        state['version'] = Version(state['version'])
        state['command']['version'] = Version(str(state['command']['version']))

        state['tests'] = [self.test_spec(**x) for x in state['tests']]
        state['node_class'] = state['class']
        del state['class']

        interface_class = fastr.interfaces[state['interface'].get('class', 'FastrInterface')]
        if 'id' not in state['interface']:
            state['interface']['id'] = '{}_{}_interface'.format(state['id'], state['command']['version'])

        if interface_class.status != PluginState.loaded:
            raise exceptions.FastrPluginNotLoaded('Required InterfacePlugin {} was not loaded properly (status {})'.format(interface_class.id,
                                                                                                                           interface_class.status))

        state['interface'] = interface_class.createobj(state['interface'])
        self.__dict__.update(state)
        self._target = None

        # TODO: set interface link

    @property
    def path(self):
        """
        The path of the directory in which the tool definition file was
        located.
        """
        return os.path.dirname(self.filename)

    @property
    def command_version(self):
        return self.command['version']

    def test(self):
        """
        Run the tests for this tool
        """
        for test in self.tests:
            self._test(test)

    def _test(self, test):
        """
        Run a single test for this tool

        :param test: the test to run
        """
        # Get data for the test
        print('Self filename: {}'.format(self.filename))
        print('Self path: {}'.format(self.path))
        test_input = {k: tuple(os.path.join(self.path, 'testdata', 'input', x) for x in v) for k, v in test.input.items()}
        test_command = test.command
        test_output = {k: tuple(os.path.join(self.path, 'testdata', 'output', x) for x in v) for k, v in test.output.items()}

        fastr.log.info("Test Input: {}".format(test_input))
        fastr.log.info("Test Command: {}".format(test_command))
        fastr.log.info("Test Input: {}".format(test_output))




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package contains the base class and meta class for all serializable
objects in the Fastr system.
"""

import gzip
import json
import marshal
import os
import pickle
from abc import abstractmethod

from jsonschema.exceptions import ValidationError

import fastr
import fastr.exceptions as exceptions
from fastr.utils import xmltodict, jsonschemaparser


class PassThroughSerializer(object):
    @staticmethod
    def dumps(data):
        return data

    @staticmethod
    def loads(data):
        return data


class Serializable(object):
    """
    Superclass for all classes that can be serialized.
    """
    dumpfuncs = {
        'dict': PassThroughSerializer,
        'json': json,
        'pickle': pickle,
        'marshall': marshal,
        'xml': xmltodict,
    }

    SERIALIZERS = {}

    @classmethod
    def get_serializer(cls, filename=None):
        if filename is None:
            if hasattr(cls, '__dataschemafile__'):
                filename = cls.__dataschemafile__
            else:
                raise exceptions.FastrValueError('Cannot get serializer for class without dataschemafile specified!')
        if not os.path.exists(filename):
            filename = os.path.join(fastr.config.schemadir, filename)

        if filename not in cls.SERIALIZERS:
            try:
                cls.SERIALIZERS[filename] = jsonschemaparser.getblueprinter(filename)
            except IOError as err:
                fastr.log.warning('Serializable class {} cannot read schemafile {}: {}'.format(
                    cls.__name__, filename, err.strerror))

        return cls.SERIALIZERS[filename]

    @abstractmethod
    def __getstate__(self):
        raise exceptions.FastrNotImplementedError('For a Serializable the __getstate__ has to be set by the subjc')

    @classmethod
    def _unserialize(cls, doc, network=None):
        """
        Classmethod that returns an object constructed based on the
        str/dict (or OrderedDict) representing the object

        :param dict doc: the state of the object to create
        :param network: the network the object will belong to
        :type network: :py:class:`Network <fastr.core.network.Network>`
        :return: newly created object
        """
        return cls.createobj(doc, network)

    @classmethod
    def createobj(cls, state, _=None):
        """
        Create object function for generic objects

        :param cls: The class to create
        :param state: The state to use to create the Link
        :param network: the parent Network
        :return: newly created Link
        """
        # Assume a state
        if hasattr(cls, '__base__'):
            # We just check if __base__ existed, but pylint missed that
            # pylint: disable=no-member
            base = cls.__base__
        else:
            raise exceptions.FastrNotImplementedError('Cannot parse old-style classes!')

        if base is object:
            obj = object.__new__(cls)
        else:
            obj = base.__new__(cls)

        if hasattr(obj, '__setstate__'):
            # We just check if __setstate__ existed, but pylint missed that
            # pylint: disable=no-member,maybe-no-member
            obj.__setstate__(state)
        else:
            obj.__dict__.update(state)

        return obj

    def _serialize(self):
        """
        Method that returns a dict/str/list structure with the data to rebuild
        the object.

        :returns: serialized representation of object
        """
        return self.__getstate__()

    def dump(self, file_handle, method='json', **kwargs):
        """
        Dump the object to a file like object.

        :param file_handle: file descriptor to write the data to
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        """
        self._dump(self._serialize(), file_handle, method, **kwargs)

    def dumps(self, method='json', **kwargs):
        """
        Dump the object to a string

        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: serialization string
        :rtype: str
        """
        return self._dumps(self._serialize(), method, **kwargs)

    def dumpf(self, path, method=None, **kwargs):
        """
        Dump the object to a file

        :param path: path where to write the file
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer

        .. note::
            The dumpf function can determine the method based on the desired
            output filename. Also, if the filename ends with .gz it will
            continue search for another extension (so .json.gz could be found)
            and will then compress the result with gzip.
        """
        self._dumpf(self._serialize(), path, method, **kwargs)

    @classmethod
    def load(cls, file_handle, method=None, network=None, **kwargs):
        """
        Load the object from a file-like object

        :param cls: class of the object
        :param file_handle: file descriptor to write the data to
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param network: network in which to place the loaded object
        :param kwargs: extra arguments passed to the final serializer
        :return: newly created object

        .. warning::
            Unlike the loadf functions, this function does not automatically
            detect gzip compression. You read a gzip using the gzip.open
            method, but not but simply opening a stream and hopeing this
            function will function.
        """
        doc = cls._load(file_handle, method, **kwargs)
        return cls._unserialize(doc, network)

    @classmethod
    def loads(cls, string, method=None, network=None, **kwargs):
        """
        Load the object from a string

        :param cls: class of the object
        :param str string: the string containing the serialized data
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param network: network in which to place the loaded object
        :param kwargs: extra arguments passed to the final serializer
        :return: newly created object
        """
        doc = cls._loads(string, method, **kwargs)
        return cls._unserialize(doc, network)

    @classmethod
    def loadf(cls, path, method=None, network=None, **kwargs):
        """
        Load the object from a file

        :param cls: class of the object
        :param path: path where to write the file
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param network: network in which to place the loaded object
        :param kwargs: extra arguments passed to the final serializer
        :return: newly created object

        .. note::
            The loadf function can determine the method of loading based on the
            filename. Also it can automatically determine whether a file is
            gzipped.
        """
        doc = cls._loadf(path, method, **kwargs)
        obj = cls._unserialize(doc, network)
        obj.filename = path
        return obj

    @classmethod
    def _dump(cls, obj, file_handle, method='json', **kwargs):
        """
        Helper function that calls the dump function of the desired final
        serialized based on the specified method.

        :param cls: class of the object
        :param obj: object to serialize
        :param file_handle: file descriptor to write the data to
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: result of the selected ``dump`` function
        """
        if method in cls.dumpfuncs:
            dump_func = cls.dumpfuncs[method].dump
        else:
            message = 'Dump module/class {} not found in _dumpfuns!'.format(method)
            fastr.log.error(message)
            raise exceptions.FastrSerializationMethodError(message)

        return dump_func(obj, file_handle, **kwargs)

    @classmethod
    def _dumps(cls, obj, method='json', **kwargs):
        """
        Helper function that calls the dumps function of the desired final
        serialized based on the specified method.

        :param cls: class of the object
        :param obj: object to serialize
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: result of the selected ``dump`` function
        """
        if method in cls.dumpfuncs:
            dumps_func = cls.dumpfuncs[method].dumps
        else:
            message = 'Dump module/class {} not found in _dumpfuns!'.format(method)
            fastr.log.error(message)
            raise exceptions.FastrSerializationMethodError(message)

        return dumps_func(obj, **kwargs)

    @classmethod
    def _dumpf(cls, obj, filename, method=None, **kwargs):
        """
        Helper function that calls the dump function of the desired final
        serialized based on the specified method. Then it writes the result
        to the specified file.

        :param cls: class of the object
        :param obj: object to serialize
        :param path: path where to write the file
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer

        .. note::
            The dumpf function can determine the method based on the desired
            output filename. Also, if the filename ends with .gz it will
            continue search for another extension (so .json.gz could be found)
            and will then compress the result with gzip.
        """
        if filename.endswith('.gz'):
            compress = True
            ext = os.path.splitext(filename[:-3])[1]
        else:
            compress = False
            ext = os.path.splitext(filename)[1]

        if method is None:
            if ext[1:] in cls.dumpfuncs:
                fastr.log.debug('Found extension as viable dump method, selecting {}'.format(ext[1:]))
                method = ext[1:]
            else:
                message = 'Could not determine dump method based on extension ({})'.format(ext)
                fastr.log.error(message)
                raise exceptions.FastrSerializationMethodError(message)

        if compress:
            with gzip.open(filename, 'w') as fout:
                cls._dump(obj, fout, method, **kwargs)
        else:
            with open(filename, 'w') as fout:
                cls._dump(obj, fout, method, **kwargs)

    @classmethod
    def _loads(cls, string, method=None, **kwargs):
        """
        Helper function to select the correct method and load the data from
        a string.

        :param cls: class of object
        :param str string: the string containing the serialized data
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: the data loaded from the file

        .. note::
            If no method is given, the function will attempt each known
            loaded and will use the first one that does not raise
            exceptions.
        """
        if method is None:
            # get all available methods, make sure json, pickle and xml are first
            methods = cls.dumpfuncs.keys()
            methods.remove('json')
            methods.remove('pickle')
            methods.remove('xml')
            methods = ['json', 'pickle', 'xml'] + methods

            for method in methods:
                try:
                    doc = cls._loads(string, method, **kwargs)
                    return doc
                except ValidationError:
                    raise
                except (ValueError, KeyError):
                    pass  # Known possible errors with json/pickle/xml parsing
                except Exception:
                    raise

            raise exceptions.FastrSerializationMethodError('Could not load string, no suitable reader found')

        if method in cls.dumpfuncs:
            loads_func = cls.dumpfuncs[method].loads
        else:
            message = 'Dump module/class not found in _dumpfuns!'
            fastr.log.error(message)
            raise exceptions.FastrSerializationMethodError(message)

        doc = loads_func(string, **kwargs)
        if hasattr(cls, '__dataschemafile__'):
            # We just check if __dataschemafile__ existed, but pylint missed that
            # pylint: disable=no-member
            doc = cls.get_serializer().instantiate(doc)

        return doc

    @classmethod
    def _load(cls, file_object, method=None, **kwargs):
        """
        Helper function to select the correct method and load the data from
        a file-like object.

        :param cls: class of object
        :param file_object: file descriptor to write the data to
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: the data loaded from the file

        .. warning::
            Unlike the loadf functions, this function does not automatically
            detect gzip compression. You read a gzip using the gzip.open
            method, but not but simply opening a stream and hopeing this
            function will function.
        """
        return cls._loads(file_object.read(), method, **kwargs)

    @classmethod
    def _loadf(cls, path, method=None, **kwargs):
        """
        Helper function to select the correct method and load the data from
        a file.

        :param cls: class of object
        :param path: path of the file to load
        :param str method: method of final serialization to use
                           (e.g. json, xml, pickle)
        :param kwargs: extra arguments passed to the final serializer
        :return: the data loaded from the file

        .. note::
            The loadf function can determine the method of loading based on the
            filename.  Also it can automatically determine whether a file is
            gzipped.
        """
        # Try to guess the method based on the extension
        if method is None:
            # get all available methods, make sure json, pickle and xml are first
            methods = cls.dumpfuncs.keys()

            for potential_method in methods:
                if path.endswith('.{}'.format(method)) or path.endswith('.{}.gz'.format(method)):
                    method = potential_method
                    break

        if not os.path.isfile(path):
            message = 'File not found ({})'.format(path)
            fastr.log.error(message)
            raise exceptions.FastrIOError(message)

        # Try if the file is GZipped
        try:
            with gzip.open(path, 'r') as fin:
                data = cls._load(fin, method, **kwargs)
                return data
        except IOError:
            pass

        # Treat as uncompressed file
        try:
            with open(path, 'r') as fin:
                data = cls._load(fin, method, **kwargs)
        except ValueError:
            #raise exceptions.FastrSerializationMethodError('Load failed! Could not find a suitable reader for {}.'.format(path))
            raise

        return data

    @classmethod
    def _unicode_convert(cls, input_):
        """
        Helper function to convert unicode instances in a data structure to str

        :param cls: class of object
        :param input_: the data structure to convert
        :return: converted version of the data
        """
        if isinstance(input_, dict):
            return {cls._unicode_convert(key): cls._unicode_convert(value) for key, value in input_.iteritems()}
        elif isinstance(input_, list):
            return [cls._unicode_convert(element) for element in input_]
        elif isinstance(input_, unicode):
            return input_.encode('utf-8')
        else:
            return input_



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from fastr.core.version import Version
from fastr.datatypes import ValueType


class String(ValueType):
    description = 'A simple string value'

    def _validate(self):
        """
        Validate the value of the DataType.

        :return: flag indicating validity of the String
        :rtype: bool
        """
        if self.value is None:
            return False
        elif isinstance(self.value, str):
            return True
        else:
            return False



#!/usr/bin/env python
import os
import sys
import textwrap
from fastr.utils.rest_generation import create_rest_table


if hasattr(sys, 'real_prefix'):
    print('[generate_plugins.py] Inside virtual env: {}'.format(sys.prefix))
else:
    print('[generate_plugins.py] Not inside a virtual env!')

# Add the fastr top level directory for importing without an install
#fastr_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../'))
#print('Source fastr from: {}'.format(fastr_dir))
#sys.path = [fastr_dir] + sys.path

import fastr


def generate_plugins(plugin_header, plugin_manager, plugin_type_name):
    plugin_type_id = plugin_type_name.lower()

    # The initial document
    doc = [plugin_header, None]
    table = []

    # Extract data from all IOPlugins
    print('[generate_plugins.py] Found Plugins: {}'.format(plugin_manager.keys()))
    for scheme, plugin in sorted(plugin_manager.items()):
        if plugin.instantiate:
            plugin = type(plugin)

        plugin_name = plugin.__name__
        underline = '^' * len(plugin_name)
        refname = '.. _{}-{}:'.format(plugin_type_id, plugin_name)
        docstring = plugin.__doc__
        if docstring is None:
            docstring = 'NOT DOCUMENTED!'

        docstring = textwrap.dedent(docstring)

        if len(plugin.configuration_fields) > 0:
            fields = plugin.configuration_fields

            field_names = fields.keys()
            field_types, field_defaults, field_description = zip(*fields.values())

            # Convert fields to str
            field_types = [x.__name__ for x in field_types]
            field_defaults = [repr(x) for x in field_defaults]
            field_description = list(field_description)

            docstring += "\n\n\n**Configuration fields**\n\n"
            docstring += create_rest_table(data=[field_names,
                                                   field_types,
                                                   field_description,
                                                   field_defaults],
                                             headers=['name',
                                                      'type',
                                                      'description',
                                                      'default'])

        table.append((scheme, plugin_name))
        doc.append('{}\n\n{}\n{}\n\n{}\n'.format(refname,
                                                 plugin_name,
                                                 underline,
                                                 docstring.strip()))

    # Make links for plugins
    table = [(x[0], ':ref:`{0} <{1}-{0}>`'.format(x[1], plugin_type_id)) for x in table]

    table = create_rest_table(
        data=zip(*table),
        headers=['scheme', ':py:class:`{name} <{mod}.{name}>`'.format(name=plugin_type_name,
                                                                      mod='fastr.plugins')]
    )

    doc[1] = table + '\n'

    # Join everything together
    doc = '\n'.join(doc)
    filename = os.path.join(os.path.dirname(__file__), 'fastr.ref.{}s.rst'.format(plugin_type_name.lower()))
    print('[generate_plugins.py] Writing {} reference to {} ({})'.format(plugin_type_name, filename, os.path.abspath(filename)))
    with open(filename, 'w') as output:
        output.write(doc)

    return filename


def find_parents(cls):
    if isinstance(cls, fastr.core.baseplugin.Plugin):

        cls = type(cls)
    if not issubclass(cls, fastr.core.baseplugin.Plugin):
        raise ValueError('Not a plugin! ({})'.format(cls))

    if fastr.core.baseplugin.Plugin in cls.__bases__:
        return [cls]
    else:
        return [y for x in cls.__bases__ for y in find_parents(x)]


def find_plugin_types():
    plugins = [base for plugin in fastr.plugin_manager.values() for base in find_parents(plugin)]
    return sorted(set(plugins), key=lambda x: x.id)


def generate_all():
    print('[generate_plugins.py] Start generating plugin references')
    plugin_types = find_plugin_types()
    files = []

    for plugin_type in plugin_types:
        doc = plugin_type.__doc__ or 'NOT DOCUMENTED!'
        header = """\
.. _{id}-ref:

{ID} Reference
{underline}

{docstring}
        """.format(
            id=plugin_type.id.lower(),
            underline='-' * (len(plugin_type.id) + 10),
            ID=plugin_type.id,
            docstring=textwrap.dedent(doc)
        )

        print('[generate_plugins.py] Start generating {} reference'.format(plugin_type.id))
        filename = generate_plugins(header,
                                    {k: v for k, v in fastr.plugin_manager.items() if isinstance(v, plugin_type) or issubclass(v, plugin_type)},
                                     plugin_type.id)
        files.append(filename)
        print('[generate_plugins.py] Finished generating {} reference'.format(plugin_type.id))

    filename = os.path.join(os.path.dirname(__file__), 'fastr.reference.rst')
    reference_header = os.path.join(os.path.dirname(__file__), 'static', 'reference.rst')

    with open(filename, 'w') as fh_out:
        with open(reference_header, 'r') as fh_in:
            fh_out.write(fh_in.read())

        for filepath in files:
            filepath = os.path.basename(filepath)
            fh_out.write('\n.. include:: {}\n'.format(filepath))


if __name__ == '__main__':
    generate_all()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The datatypes module holds all DataTypes generated by fastr and all the base
classes for these datatypes.
"""

# Empty module to be populated by the fastr.core.datatypemanager.DataTypeManager
from abc import abstractmethod, abstractproperty
import os
import sys
import traceback
import urllib
import urlparse

from fastr import exceptions as exceptions
import fastr
from fastr.core.baseplugin import BasePlugin, PluginState
from fastr.core.version import Version
from fastr.data import url
from fastr.utils.checksum import md5_checksum, hashsum
from fastr.utils.classproperty import classproperty
from fastr.utils import iohelpers


class BaseDataType(BasePlugin):
    """
    The base class for all datatypes in the fastr type system.
    """
    filename = __file__

    #: Version of the DataType definition
    version = Version('1.0')

    #: Description of the DataType
    description = ''

    #: Extension related to the Type
    extension = None

    # DataTypes do not need to be loaded, so they are always Loaded
    _status = (PluginState.loaded, 'DataTypes are always loaded', '')

    @abstractmethod
    def __init__(self, value=None, format_=None):
        """
        The BaseDataType constructor.

        :param value: value to assign to the new BaseDataType object
        :param format_: the format used for the ValueType
        :return: new BaseDataType object
        :raises FastrDataTypeNotInstantiableError: if not subclassed
        :raises FastrNotImplementedError: if *id*, *name*, *version* or *description* is None
        """
        super(BaseDataType, self).__init__()

        if self.__class__ is BaseDataType:
            raise exceptions.FastrDataTypeNotInstantiableError('BaseDataType not instantiable')

        # Check if id, name and version are set properly
        if self.id is None or self.name is None or self.version is None or self.description is None:
            raise exceptions.FastrNotImplementedError(('Subclasses of BaseDataType need to'
                                                       ' have id ({}), name ({}), version ({})'
                                                       ' and description ({}) fields set!').format(self.id,
                                                                                                   self.name,
                                                                                                   self.version,
                                                                                                   self.description))

        self._value = None
        self.value = value
        self.format = format_

        if not self.valid:
            fastr.log.debug("'{}' is not a valid {}".format(self._value, self.id))

    def __repr__(self):
        """
        Returns string representation of the BaseDataType

        :return: string represenation
        :rtype: str
        """
        return "<{}: {}>".format(self.id, repr(self.value))

    def __str__(self):
        """
        Returns the string version of the BaseDataType

        :return: string version
        :rtype: str
        """
        return str(self.value)

    def __eq__(self, other):
        """
        Test the equality of two DataType objects

        :parameter DataType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        if type(self) is not type(other):
            return NotImplemented

        return self.value == other.value

    def __ne__(self, other):
        """
        Test if two objects are not equal. This is by default done by negating
        the __eq__ operator

        :parameter DataType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        return not self.__eq__(other)

    def __getstate__(self):
        return (type(self).id, self.value, self.format)

    def __setstate__(self, state):
        if self.id != state[0]:
            raise exceptions.FastrValueError('Unvalid state for {}, state is for type {}'.format(self.id, state[0]))

        self.value = state[1]
        self.format = state[2]

    @classproperty
    def id(cls):
        """
        Internal id used for the DataType
        """
        return cls.__name__

    @classproperty
    def fullid(cls):
        """
        The full fastr id of the DataType
        """
        return '{}/{}'.format(cls.parent.fullid, cls.id)

    @classproperty
    def name(cls):
        """
        Display friendly name of the DataType
        """
        return cls.id

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        return self._value

    @property
    def raw_value(self):
        """
        The raw value of object instantiation of this DataType. For datatypes
        that override value (like Deferred) this is the way to access the
        _value field.
        """
        result = self._value

        if isinstance(result, basestring):
            if result.startswith('val://'):
                result = result[6:]
            result = urllib.quote_plus(result)

        return result

    @property
    def data_uri(self):
        if url.isurl(self._value):
            return self._value
        else:
            return 'fastr://data/constant/{}'.format(self._value)

    @value.setter
    def value(self, value):
        """
        Setter function for value property
        """
        if isinstance(value, type(self)):
            self._value = value.value
            self.format = value.format
        else:
            self._value = value

    @classproperty
    def parent(self):
        """
        The parent container of the DataType
        """
        return fastr.typelist

    @property
    def parsed_value(self):
        """
        The parsed value of object instantiation of this DataType.
        """
        return self._value

    @classmethod
    def test(cls):
        """
        Define the test for the BasePluginManager. Make sure we are not one of the
        base classes
        """
        if cls in [BaseDataType, DataType, TypeGroup, EnumType, ValueType, URLType]:
            raise exceptions.FastrTypeError('Cannot use a basic datatype {} as a plugin!'.format(cls))

    @property
    def valid(self):
        """
        A boolean flag that indicates weather or not the value assigned to this
        DataType is valid. This property is generally overwritten by
        implementation of specific DataTypes.
        """
        # Avoid errors in the validation testing (this is external code)
        # so we use a broad except on purpose
        # pylint: disable=broad-except
        try:
            return self._validate()
        except Exception:
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            fastr.log.warning('Could not validate {}: encountered exception ({}) during execution:\n{}'.format(repr(self), exc_type.__name__, exc_info))
            raise
            return False

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        # This function doesn't use self, but is intended to be potentially
        # overwritten by the subclasses.
        # pyline: disbale=no-self-use
        return True

    def checksum(self):
        """
        Generate a checksum for the value of this DataType

        :return: the checksum of the value
        :rtype: str
        """
        return hashsum(self.value)

    @classmethod
    def isinstance(cls, value):
        """
        Indicate whether value is an instance for this DataType.

        :return: the flag indicating the value is of this DataType
        :rtype: bool
        """
        return isinstance(value, cls)


class DataType(BaseDataType):
    """
    This class is the base class for all DataTypes that can hold a value.
    """
    @abstractmethod
    def __init__(self, value=None, format_=None):
        """
        The DataType constructor.

        :param value: value to assign to the new DataType object
        :param format: the format used for the ValueType
        :return: new DataType object
        """
        super(DataType, self).__init__(value, format_)

    def action(self, name):
        """
        This function can be overwritten by subclasses to implement certain
        action that should be performed. For example, the *Directory* DataType
        has an action *ensure*. This method makes sure the Directory exists.
        A Tool can indicate an action that should be called for an Output which
        will be called before execution.

        :param str name: name of the action to execute
        :return: None
        """
        if name is not None:
            fastr.log.warning("unknown action '{}' for DataType {}".format(name, self.id))


class TypeGroup(BaseDataType):
    """
    The TypeGroup is a special DataType that does not hold a value of its own
    but is used to group a number of DataTypes. For example ITK has a list of
    supported file formats that all tools build on ITK support. A group can
    be used to conveniently specify this in multiple Tools that use the same
    set DataTypes.
    """

    def __new__(cls, value=None, format_=None):
        """
        Instantiate a TypeGroup. This will for match the value to the best
        matching type and instantiate that. Not that the returned object
        will not be of type TypeGroup but one of the TypeGroup members.
        """
        # Avoid casting values that are already in a member type
        if any(isinstance(value, x) for x in cls.members):
            return value

        matching_type = fastr.typelist.guess_type(value, options=cls)

        # Only continue if we have value urls
        if isinstance(value, (str, unicode)) and value.startswith('val://'):
            if matching_type is None:
                # Just try a simple match rather than guessing
                matching_type = fastr.typelist.match_types(cls, type(value))

            if matching_type is None:
                # Just try a simple match rather than guessing
                matching_type = fastr.typelist.match_types(cls)

        if matching_type is None:
            raise exceptions.FastrValueError('Cannot matching value {} [{}] to any of {}'.format(value,
                                                                                                 type(value).__name__,
                                                                                                 cls.members))

        return matching_type(value, format_)

    def __init__(self, value=None):
        """
        Dummy constructor. TypeGroups are not instantiable and cannot hold a
        value of its own.

        :raises FastrDataTypeNotInstantiableError: if called
        """
        # All type groups are per definition not instantiable
        # pylint: disable=super-init-not-called
        raise exceptions.FastrDataTypeNotInstantiableError('TypeGroups are not instantiable')

    @abstractproperty
    def _members(self):
        """
        The id of the members, this should be set in the subclass. It should
        be a frozen set of str.
        """
        _members = frozenset()

    @classproperty
    def members(cls):
        """
        The members of the TypeGroup
        """
        members = frozenset(fastr.typelist[x] for x in cls._members if x in fastr.typelist)
        members = [x.members if issubclass(x, TypeGroup) else (x,) for x in members]
        return frozenset(x for y in members for x in y)

    @classmethod
    def isinstance(cls, value):
        return any(x.isinstance(value) for x in cls.members)


class AnyType(TypeGroup):
    """
    Special Datatype in fastr that is a TypeGroup with all known DataTypes as
    its members.
    """
    @classproperty
    def _members(cls):
        """
        A "class-poperty" that gives a list of the ids of all currently loaded DataTypes
        """
        return frozenset(x.id for x in fastr.typelist.values() if issubclass(x, DataType))

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_members = ['  - {}'.format(member) for member in cls.members]
        return """
TypeGroup {id}
{name} ({id}) is a group of consisting of all DataTypes known by fastr, currently:
{members}
""".strip().format(id='AnyType', name='AnyType', members='\n'.join(disp_members))


class AnyFile(TypeGroup):
    """
    Special Datatype in fastr that is a TypeGroup with all known DataTypes as
    its members.
    """

    @classproperty
    def _members(cls):
        """
        A "class-poperty" that gives a list of the ids of all currently loaded DataTypes
        """
        return frozenset(x.id for x in fastr.typelist.values() if issubclass(x, URLType))

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_members = ['  - {}'.format(member) for member in cls.members]
        return """
TypeGroup {id}
{name} ({id}) is a group of consisting of all URLTypes known by fastr, currently:
{members}
""".strip().format(id='AnyFile', name='AnyFile', members='\n'.join(disp_members))


class EnumType(DataType):
    """
    The EnumType is the base for DataTypes that can have a value which is an
    option from a predefined set of possibilities (similar to an enum type
    in many programming languages).
    """
    #: Enums always have version 1.0
    version = Version('1.0')
    _options = frozenset()

    def __init__(self, value=None, format_=None):
        """
        The EnumType constructor.

        :param value: value to assign to the new EnumType object
        :param format: the format used for the ValueType
        :return: new EnumType object
        :raises FastrDataTypeNotInstantiableError: if not subclassed
        """
        super(EnumType, self).__init__(value, format_)

        if self.__class__ is EnumType:
            raise exceptions.FastrDataTypeNotInstantiableError('EnumType is not instantiable')

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_options = ['  - "{}"'.format(option) for option in cls.options]
        return """
{name} ({id}) is a enumerate type with options:
{options}

{name} can take the value of any of the option, but any other value is considered invalid.
""".strip().format(id=cls.id, name=cls.name, options='\n'.join(disp_options))

    @classproperty
    def options(cls):
        """
        A frozenset holding the options that the value of the EnumType object can have.

        :return: the options the value can hold
        :rtype: frozenset
        """
        return cls._options

    def _validate(self):
        return self._value in self._options


class ValueType(DataType):
    """
    The ValueType is the base for DataTypes that hold simple values (not an
    EnumType and not a file/URL). The values is generally represented by a
    string.
    """

    def __init__(self, value=None, format_=None):
        """
        The ValueType constructor

        :param value: value to assign to the new ValueType
        :param format: the format used for the ValueType
        :return: new ValueType object
        """
        super(ValueType, self).__init__(value, format_)


class URLType(DataType):
    """
    The URLType is the base for DataTypes that point to a resource somewhere
    else (typically a filesystem). The true value is actually the resource
    referenced by the value in this object.
    """

    def __init__(self, value=None, format_=None):
        """
        The URLType constructor

        :param value: value to assign to the new URLType
        :param format: the format used for the ValueType
        :return: new URLType object
        """
        super(URLType, self).__init__(value, format_)

    def __eq__(self, other):
        """
        Test the equality of two DataType objects

        :parameter URLType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        if type(self) is not type(other):
            return NotImplemented

        return self.checksum() == other.checksum()

    def checksum(self):
        """
        Return the checksum of this URL type

        :return: checksum string
        :rtype: str
        """
        contents = self.content(self.parsed_value)
        return md5_checksum(contents)

    @property
    def parsed_value(self):
        """
        The parsed value of object instantiation of this DataType.
        """
        if url.isurl(self.value):
            parsed_url = urlparse.urlparse(self.value)

            if parsed_url.scheme == 'val':
                datafile = os.path.join(fastr.config.mounts[parsed_url.netloc], os.path.normpath(parsed_url.path[1:]))
                query = urlparse.parse_qs(parsed_url.query)

                # Open Job file
                data = iohelpers.load_gpickle(datafile)

                # Attempt to extract data
                try:
                    outputname = query['outputname'][0]
                    cardinality_nr = int(query['nr'][0])
                    if 'sampleid' in query:
                        sample_id = query['sampleid'][0]
                        value = data.output_data[outputname][sample_id][cardinality_nr]
                    else:
                        value = data.output_data[outputname][cardinality_nr]
                except (IndexError, KeyError) as exception:
                    fastr.log.debug('Output data for query: {}'.format(data.output_data))
                    message = 'Could not get value from {}, encountered {}: {}'.format(self.value, type(exception).__name__, exception.message)
                    raise exceptions.FastrKeyError(message)

                if value.startswith('vfs://'):
                    return fastr.vfs.url_to_path(value)
                else:
                    return value

            elif parsed_url.scheme == 'vfs':
                return fastr.vfs.url_to_path(self.value)
            else:
                raise exceptions.FastrValueError('Cannot get parsed value for (non-vfs, non-val) url: {} (scheme {})'.format(self.value, parsed_url.scheme))
        else:
            return self.value

    @classmethod
    def content(cls, inval, outval=None):
        """
        Give the contents of a URLType, this is generally useful for filetypes
        that consists of multiple files (e.g. AnalyzeImageFile, DICOM). The
        value will indicate the main file, and the contents function can
        determine all files that form a single data value.

        :param inval: a value to figure out contents for this type
        :param outval: the place where the copy should point to
        :return: a list of all files part of the value (e.g. header and data file)
        :rtype: list
        """
        if outval is not None:
            return [(inval, outval)]
        else:
            return [inval]

    @property
    def valid(self):
        """
        A boolean flag that indicates weather or not the value assigned to this
        DataType is valid. This property is generally overwritten by
        implementation of specific DataTypes.
        """
        if not isinstance(self.value, str):
            return False

        return super(URLType, self).valid

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        try:
            contents = self.content(self.parsed_value)

            for content in contents:
                if not os.path.exists(content):
                    return False
            return True
        except (TypeError, ValueError, IOError):
            return False


class Deferred(DataType):
    def __init__(self, value=None, format_=None):
        """
        The Deferred constructor.

        :param value: value to assign to the new DataType object
        :param format: This is ignore but here for compatibility
        :return: new Deferred object
        """
        self._value = value
        self._data = None

    def __repr__(self):
        """
        Returns string representation of the BaseDataType

        :return: string represenation
        :rtype: str
        """
        if self.target is None:
            return "<{}: {}>".format(self.id, repr(self._value))
        else:
            return "<{}: {}>".format(self.id, repr(self.target))

    def __getstate__(self):
        return ('Deferred', self._value)

    def __setstate__(self, state):
        if self.id != state[0]:
            raise exceptions.FastrValueError('Unvalid state for {}, state is for type {}'.format(self.id, state[0]))

        self._value = state[1]

    @classmethod
    def lookup(cls, value):
        """
        Look up the deferred target and return that object

        :param: value
        :return: The value the deferred points to
        :rtype: DataType
        :raises FastrKeyError: if the deferred is not available (yet)
        :raises FastrValueError: if the value is not a valid deferrred url
        """
        parsed_url = urlparse.urlparse(value)

        if parsed_url.scheme == 'val':
            datafile = os.path.join(fastr.config.mounts[parsed_url.netloc], os.path.normpath(parsed_url.path[1:]))
            query = urlparse.parse_qs(parsed_url.query)

            # Open Job file
            data = iohelpers.load_gpickle(datafile)

            # Get provenace
            provenance = data.provenance

            # Attempt to extract data
            try:
                outputname = query['outputname'][0]
                cardinality_nr = int(query['nr'][0])
                if 'sampleid' in query:
                    sample_id = query['sampleid'][0]
                    value = data.output_data[outputname][sample_id][cardinality_nr]
                else:
                    value = data.output_data[outputname][cardinality_nr]
            except (IndexError, KeyError) as exception:
                fastr.log.debug('Output data for query: {}'.format(data.output_data))
                message = 'Could not get value from {}, encountered {}: {}'.format(value, type(exception).__name__, exception.message)
                raise exceptions.FastrKeyError(message)
        else:
            raise exceptions.FastrValueError('Cannot lookup value {}, wrong url scheme'.format(value))

        if isinstance(value, Deferred):
            value, provenance, data = value._get_data()

        return value, provenance, data

    def _get_data(self):
        """
        Get and cache the data (target / provenance) of the
        target.

        :return: tuple containing (target, provenance)
        :rtype: tuple
        """
        if self._data is None:
            self._data = self.lookup(self._value)

        return self._data

    @property
    def target(self):
        """
        Target object for this deferred.

        :raises FastrKeyError: if the deferred is not available (yet)
        :raises FastrValueError: if the value is not a valid deferrred url
        """
        return self._get_data()[0]

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        try:
            target = self.target
        except exceptions.FastrKeyError:
            return None

        return target.value

    @property
    def provenance(self):
        return self._get_data()[1]

    @property
    def job(self):
        try:
            return self._get_data()[2]
        except exceptions.FastrValueError:
            return None

    @property
    def data_uri(self):
        if self._value.startswith('val://'):
            val_uri = urlparse.urlparse(self._value)
            val_query = urlparse.parse_qs(val_uri.query)
            sample_id = val_query['sampleid'][0] if 'sampleid' in val_query else self.job.sample_id

            return 'fastr://data/job/{jobid}/output/{outputid}/sample/{sample}/cardinality/{cardinality}'.format(
                jobid=self.job.jobid,
                outputid=val_query['outputname'][0],
                sample=sample_id,
                cardinality=val_query['nr'][0]
            )
        else:
            return super(Deferred, self).data_uri

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        # This function doesn't use self, but is intended to be potentially
        # overwritten by the subclasses.
        # pyline: disbale=no-self-use
        try:
            target = self.target
        except exceptions.FastrKeyError:
            return False

        return target.valid

    def checksum(self):
        """
        Generate a checksum for the value of this DataType

        :return: the checksum of the value
        :rtype: str
        """
        try:
            target = self.target
        except exceptions.FastrKeyError:
            return hashsum('__FASTR_NOT_AVAILABLE_HASH__')

        return target.checksum()


def fastr_isinstance(obj, datatype):
    """
    Check if an object is of a specific datatype.

    :param obj: Object to inspect
    :param datatype: The datatype(s) to check
    :type datatype: tuple, BaseDataType
    :return: flag indicating object is of datatype
    :rtype: bool
    """
    if not isinstance(datatype, tuple):
        datatype = datatype,

    for dtype in datatype:
        if issubclass(dtype, TypeGroup):
            if type(obj) in dtype.members:
                return True
        elif isinstance(obj, datatype):
            return True

    return False



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Null plugin for fastr
"""

from fastr.core.ioplugin import IOPlugin


class Null(IOPlugin):
    """
    The Null plugin is create to handle ``null://`` type or URLs. These URLs
    are indicating the sink should not do anything. The data is not written to
    anywhere. Besides the scheme, the rest of the URL is ignored.
    """
    scheme = 'null'

    def __init__(self):
        # initialize the instance and register the scheme
        super(Null, self).__init__()

    def put_url(self, inpath, outurl):
        """
        Put the files to the external data store.

        :param inpath: path of the local data
        :param outurl: url to where to store the data, starts with ``file://``
        """
        return True

    def put_value(self, value, outurl):
        """
        Put the value in the external data store.

        :param value: value to store
        :param outurl: url to where to store the data, starts with ``file://``
        """
        return True



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def create_rest_table(data, headers):
    """
    Create a ReST table from data. The data should be a list of columns and the
    headers should be a list of column names.

    :param list data: List of lists/tuples representing the columns
    :param list headers: List of strings for the column names
    :returns: a string representing the table in ReST
    :rtype: str
    """
    # Data should be a list of columns
    if len(data) != len(headers):
        raise ValueError('Number of columns should match the number of header values')

    # Get size of strings per column
    sizes = [max(len(x) for x in col) for col in data]
    sizes = [max(s, len(h)) for s, h in zip(sizes, headers)]

    table_line = ' '.join('=' * s for s in sizes)
    format_base = ' '.join('{{line[{}]:<{}}}'.format(n, s) for n, s in enumerate(sizes))

    table = [
        table_line,
        format_base.format(line=headers),
        table_line
    ]

    # Iterate transposed data
    table.extend(format_base.format(line=line) for line in zip(*data))
    table.append(table_line)

    return '\n'.join(table)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import nose.tools as nt
import pickle

import fastr
from fastr import exceptions
from fastr.core.samples import SampleId, SampleIndex, SampleItem, SampleCollection

String = fastr.typelist['String']


class TestSamples():
    def setup(self):
        self.sample_id_0 = SampleId('single')
        self.sample_id_1 = SampleId('test', '123', 'test')
        self.sample_id_2 = SampleId(['built', 'from', 'list'])

        self.sampleindex_0 = SampleIndex(0)
        self.sampleindex_1 = SampleIndex(42, 4, 2)
        self.sampleindex_2 = SampleIndex([5, 3, 8])
        self.sampleindex_3 = SampleIndex(xrange(3))  # Create index from iterator
        self.sampleindex_4 = SampleIndex(0, slice(3, 5))
        self.sampleindex_5 = SampleIndex(slice(1, 6, 2), 3)
        self.sampleindex_6 = SampleIndex(slice(None, None, 2))
        self.sampleindex_7 = SampleIndex(slice(None, 9, 3))
        self.sampleindex_8 = SampleIndex(slice(None, 5, 2), slice(1, 4))

        self.sample_item_0 = SampleItem(self.sampleindex_0, self.sample_id_0, OrderedDict(), None)
        self.sample_item_1 = SampleItem(self.sampleindex_1, self.sample_id_1, OrderedDict({0: (String('val1'), String('val2'))}), set())
        self.sample_item_2 = SampleItem((0, 1), ('a', 'b'), OrderedDict({0: (String('val1'), String('val2'))}), set())

    def test_sample_id_str(self):
        nt.eq_(str(self.sample_id_0), 'single')
        nt.eq_(str(self.sample_id_1), 'test__123__test')
        nt.eq_(str(self.sample_id_2), 'built__from__list')

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleid_noniter(self):
        SampleId(None)

    @nt.raises(exceptions.FastrValueError)
    def test_sampleindex_empty(self):
        SampleIndex()

    @nt.raises(exceptions.FastrValueError)
    def test_sampleindex_emptylist(self):
        SampleIndex([])

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_noniter(self):
        SampleIndex(None)

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_wrongtype1(self):
        SampleIndex('test')

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_wrongtype2(self):
        SampleIndex(['test', 1, 2, 3])

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_wrongtype3(self):
        SampleIndex(1.0, 2.0)

    def test_sampleindex_str(self):
        nt.eq_(str(self.sampleindex_0), '(0)')
        nt.eq_(str(self.sampleindex_1), '(42, 4, 2)')
        nt.eq_(str(self.sampleindex_2), '(5, 3, 8)')
        nt.eq_(str(self.sampleindex_3), '(0, 1, 2)')
        nt.eq_(str(self.sampleindex_4), '(0, 3:5)')
        nt.eq_(str(self.sampleindex_5), '(1:6:2, 3)')
        nt.eq_(str(self.sampleindex_6), '(::2)')
        nt.eq_(str(self.sampleindex_7), '(:9:3)')
        nt.eq_(str(self.sampleindex_8), '(:5:2, 1:4)')

    def test_sampleindex_repr(self):
        nt.eq_(repr(self.sampleindex_0), '<SampleIndex (0)>')
        nt.eq_(repr(self.sampleindex_1), '<SampleIndex (42, 4, 2)>')
        nt.eq_(repr(self.sampleindex_2), '<SampleIndex (5, 3, 8)>')
        nt.eq_(repr(self.sampleindex_3), '<SampleIndex (0, 1, 2)>')
        nt.eq_(repr(self.sampleindex_4), '<SampleIndex (0, 3:5)>')
        nt.eq_(repr(self.sampleindex_5), '<SampleIndex (1:6:2, 3)>')
        nt.eq_(repr(self.sampleindex_6), '<SampleIndex (::2)>')
        nt.eq_(repr(self.sampleindex_7), '<SampleIndex (:9:3)>')
        nt.eq_(repr(self.sampleindex_8), '<SampleIndex (:5:2, 1:4)>')

    def test_sampleindex_isslice(self):
        nt.ok_(not self.sampleindex_0.isslice)
        nt.ok_(not self.sampleindex_1.isslice)
        nt.ok_(not self.sampleindex_2.isslice)
        nt.ok_(not self.sampleindex_3.isslice)
        nt.ok_(self.sampleindex_4.isslice)
        nt.ok_(self.sampleindex_5.isslice)
        nt.ok_(self.sampleindex_6.isslice)
        nt.ok_(self.sampleindex_7.isslice)
        nt.ok_(self.sampleindex_8.isslice)

    @nt.raises(exceptions.FastrValueError)
    def test_sampleindex_expand_wrong_dim(self):
        self.sampleindex_8.expand((10, 8, 9))

    def test_sampleindex_expand(self):
        # No expand used
        nt.eq_(self.sampleindex_0.expand((10,)), SampleIndex(0),)

        # Check various expands
        nt.eq_(self.sampleindex_4.expand((6, 8,)), (SampleIndex(0, 3),
                                                    SampleIndex(0, 4)))
        nt.eq_(self.sampleindex_5.expand((8, 5)), (SampleIndex(1, 3),
                                                   SampleIndex(3, 3),
                                                   SampleIndex(5, 3)))
        nt.eq_(self.sampleindex_6.expand((5,)), (SampleIndex(0),
                                                 SampleIndex(2),
                                                 SampleIndex(4)))
        nt.eq_(self.sampleindex_6.expand((6,)), (SampleIndex(0),
                                                 SampleIndex(2),
                                                 SampleIndex(4)))
        nt.eq_(self.sampleindex_6.expand((7,)), (SampleIndex(0),
                                                 SampleIndex(2),
                                                 SampleIndex(4),
                                                 SampleIndex(6)))
        nt.eq_(self.sampleindex_8.expand((7, 6)), (SampleIndex(0, 1),
                                                   SampleIndex(0, 2),
                                                   SampleIndex(0, 3),
                                                   SampleIndex(2, 1),
                                                   SampleIndex(2, 2),
                                                   SampleIndex(2, 3),
                                                   SampleIndex(4, 1),
                                                   SampleIndex(4, 2),
                                                   SampleIndex(4, 3)))

    def test_sampleindex_add(self):
        # Add sample index
        nt.eq_(self.sampleindex_0 + self.sampleindex_1, SampleIndex(0, 42, 4, 2))
        nt.eq_(self.sampleindex_2 + self.sampleindex_6, SampleIndex(5, 3, 8, slice(None, None, 2)))

        # Add element
        nt.eq_(self.sampleindex_0 + 5, SampleIndex(0, 5))
        nt.eq_(self.sampleindex_0 + slice(1, 4), SampleIndex(0, slice(1, 4)))

        # Add tuple
        nt.eq_(self.sampleindex_0 + (4, 5), SampleIndex(0, 4, 5))

    @nt.raises(TypeError)
    def test_sampleindex_add_wrong_type(self):
        self.sampleindex_0 + 1.0

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_add_wrong_type2(self):
        self.sampleindex_0 + (1.0, 2.0)

    def test_sampleindex_radd(self):
        # Radd element
        nt.eq_(4 + self.sampleindex_0, SampleIndex(4, 0))
        nt.eq_(slice(4, 8) + self.sampleindex_1, SampleIndex(slice(4, 8), 42, 4, 2))

        # Add tuple
        nt.eq_((1, 2) + self.sampleindex_0, SampleIndex(1, 2, 0))
        nt.eq_((slice(None, 3), 1) + self.sampleindex_2, SampleIndex(slice(None, 3), 1, 5, 3, 8))

    @nt.raises(TypeError)
    def test_sampleindex_radd_wrong_type(self):
        1.0 + self.sampleindex_0

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleindex_radd_wrong_type2(self):
        (1.0, 2.0) + self.sampleindex_0

    @nt.raises(exceptions.FastrTypeError)
    def test_sampleitem_wrong_jobs_type(self):
        SampleItem((0, 1), ('a', 'b'), {0: ('val1', 'val2')}, dict())

    def test_sampleitem_repr(self):
        nt.eq_(repr(self.sample_item_0), '<SampleItem index=(0), id=single>')
        nt.eq_(repr(self.sample_item_1), '<SampleItem index=(42, 4, 2), id=test__123__test>')

    def test_sampleitem_pickle(self):
        nt.eq_(self.sample_item_0, pickle.loads(pickle.dumps(self.sample_item_0)))
        nt.eq_(self.sample_item_1, pickle.loads(pickle.dumps(self.sample_item_1)))
        nt.eq_(self.sample_item_2, pickle.loads(pickle.dumps(self.sample_item_2)))
        nt.ok_(self.sample_item_0 != pickle.loads(pickle.dumps(self.sample_item_1)))
        nt.ok_(self.sample_item_0 != pickle.loads(pickle.dumps(self.sample_item_2)))
        nt.ok_(self.sample_item_1 != pickle.loads(pickle.dumps(self.sample_item_2)))

    def test_sampleitem_newargs(self):
        nt.eq_(self.sample_item_0.__getnewargs__(), (self.sampleindex_0, self.sample_id_0, OrderedDict(), set()))
        nt.eq_(self.sample_item_1.__getnewargs__(), (self.sampleindex_1, self.sample_id_1, OrderedDict({0: (String('val1'), String('val2'))}), set()))
        nt.eq_(self.sample_item_2.__getnewargs__(), (SampleIndex(0, 1), SampleId('a', 'b'), OrderedDict({0: (String('val1'), String('val2'))}), set()))

    def test_sampleitem_data(self):
        nt.eq_(self.sample_item_0.data, OrderedDict())
        nt.eq_(self.sample_item_1.data, OrderedDict({0: (String('val1'), String('val2'))}))
        nt.eq_(self.sample_item_2.data, OrderedDict({0: (String('val1'), String('val2'))}))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module that describes the interface of a Tool. It specifies how a set of
 input values will be translated to commands to be executed. This creates
 a generic interface to different ways of executing underlying software.
"""

from abc import ABCMeta, abstractmethod, abstractproperty
from collections import namedtuple

import fastr
from fastr import exceptions
from fastr.core.baseplugin import Plugin
from fastr.core.pluginmanager import PluginSubManager
from fastr.core.serializable import Serializable


# Simple containers for minial fields for the input and output specification
InputSpecBase = namedtuple('InputSpec', ['id', 'cardinality', 'datatype', 'required', 'description', 'default', 'hidden'])
OutputSpecBase = namedtuple('OutputSpec', ['id', 'cardinality', 'datatype', 'automatic', 'required', 'description', 'hidden'])


class InputSpec(InputSpecBase):
    def __new__(cls, id_, cardinality, datatype, required=False, description='', default=None, hidden=False):
        return super(InputSpec, cls).__new__(cls,
                                             id_,
                                             cardinality=cardinality,
                                             datatype=datatype,
                                             required=required,
                                             description=description,
                                             default=default,
                                             hidden=hidden)


class OutputSpec(OutputSpecBase):
    def __new__(cls, id_, cardinality, datatype, automatic=True, required=False, description='', hidden=False):
        return super(OutputSpec, cls).__new__(cls,
                                              id_,
                                              cardinality=cardinality,
                                              datatype=datatype,
                                              automatic=automatic,
                                              required=required,
                                              description=description,
                                              hidden=hidden)


class Interface(Plugin, Serializable):

    """
    Abstract base class of all Interfaces. Defines the minimal requirements for
    all Interface implementations.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def execute(self, target, payload):
        """
        Execute the interface given the a target and payload. The payload
        should have the form {'input': {'input_id_a': (value, value), 'input_id_b': (value, value)}, 'output': {'output_id_a': (value, value), 'output_id_b': (value, value)}}

        :param target: the target to call
        :param payload: the payload to use
        :returns: the result of the execution
        :rtype: (tuple of) InterfaceResult
        """

    @abstractproperty
    def expanding(self):
        """
        Indicates whether or not this Interface will result in multiple samples
        per run. If the flow is unaffected, this will be zero, if it is nonzero
        it means that number of dimension will be added to the sample array.
        """

    @abstractproperty
    def inputs(self):
        """
        OrderedDict of Inputs connected to the Interface. The format should be
        {input_id: InputSpec}.
        """

    @abstractproperty
    def outputs(self):
        """
        OrderedDict of Output connected to the Interface. The format should be
        {output_id: OutputSpec}.
        """

    @abstractmethod
    def __getstate__(self):
        """
        Retrieve the state of the Interface

        :return: the state of the object
        :rtype dict:
        """

    @abstractmethod
    def __setstate__(self, state):
        """
        Set the state of the Interface
        """

    @classmethod
    def test(cls):
        """
        Test the plugin, interfaces do not need to be tested on import
        """
        pass


class InterfaceResult(object):
    """
    The class in which Interfaces should wrap their results to be picked up by fastr
    """
    def __init__(self, result_data, log_data, payload, sample_index=None, sample_id=None):
        if ((sample_index is None and sample_id is not None) or
           (sample_index is not None and sample_id is None)):
            raise exceptions.FastrValueError('An InterfaceResult needs both'
                                             ' sample_index and sample_id to'
                                             ' be set or neither of them set.')

        self.sample_index = sample_index
        self.sample_id = sample_id
        self.result_data = result_data
        self.log_data = log_data
        self.payload = payload


class InterfacePluginManager(PluginSubManager):
    """
    Container holding all the CollectorPlugins
    """

    def __init__(self):
        """
        Create the Coll
        :param path:
        :param recursive:
        :return:
        """
        super(InterfacePluginManager, self).__init__(parent=fastr.plugin_manager,
                                                     plugin_class=self.plugin_class)

    @property
    def plugin_class(self):
        """
        The class of the Plugins in the collection
        """
        return Interface

    @property
    def _instantiate(self):
        """
        Indicate that the plugins should instantiated before stored
        """
        return False



import os.path
import re
import fastr
from fastr.data import url
from fastr.datatypes import URLType
from fastr.utils.checksum import md5_checksum, hashsum


class MetaImageFile(URLType):
    description = 'Meta Image file format'
    extension = 'mhd'

    def _validate(value):
        value = value.value

        try:
            if url.isurl(value):
                value = url.get_path_from_url(value)
            fastr.log.debug('Check if {} exists...'.format(value))
            return os.path.exists(value)
        except ValueError:
            return False

    def __eq__(self, other):
        if not isinstance(other, MetaImageFile):
            return NotImplemented

        self_header = self._parse()
        other_header = other._parse()

        self_raw = url.normurl(url.join(os.path.dirname(self.parsed_value), self_header['ElementDataFile']))
        other_raw = url.normurl(url.join(os.path.dirname(other.parsed_value), other_header['ElementDataFile']))

        # These are allowed to be different as long as their content is the same
        del self_header['ElementDataFile']
        del other_header['ElementDataFile']

        if self_header != other_header:
            return False

        if md5_checksum(self_raw) != md5_checksum(other_raw):
            return False

        return True

    def checksum(self):
        """
        Return the checksum of this MetaImageFile

        :return: checksum string
        :rtype: str
        """
        header = self._parse()
        raw = url.normurl(url.join(os.path.dirname(self.parsed_value), header['ElementDataFile']))

        del header['ElementDataFile']

        with open(raw, 'rb') as fh_in:
            result = hashsum([header, fh_in])

        return result

    def _parse(self):
        data = {}
        with open(self.parsed_value) as header_file:
            for line in header_file:
                if line.strip()[0] == '#':
                    continue

                key, value = line.split('=', 1)
                data[key.strip()] = value.strip()
        return data

    @classmethod
    def content(cls, invalue, outvalue=None):
        fastr.log.debug('Determining content of MetaImageFile from invalue "{}" and outvalue "{}"'.format(invalue, outvalue))
        if url.isurl(invalue):
            if outvalue is None or url.isurl(outvalue):
                raise ValueError('Either invalue or outvalue should be a path')

            value = outvalue
            invalue_dir = url.dirurl(invalue)
        else:
            value = invalue
            invalue_dir = os.path.dirname(invalue)

        if outvalue is not None:
            if url.isurl(outvalue):
                outvalue_dir = url.dirurl(outvalue)
            else:
                outvalue_dir = os.path.dirname(outvalue)

            contents = [(invalue, outvalue)]
        else:
            contents = [invalue]

        try:
            fastr.log.debug('Trying to open header file: {}'.format(value))
            with open(value, 'r') as fin:
                # This is dangerous for large files, but mhd headers should be
                # small so we should be fine (we can change to loop over lines)
                data = fin.read()

                # Find datafile using a multiline regexp
                match = re.search('^ElementDataFile\s+=\s+(.*)$', data, re.MULTILINE)

                # Get filename and make absolute
                datafile = match.group(1)
                if datafile[0] in '/\\':
                    message = 'Fastr does not support MetaImageFiles with absolute paths!'
                    fastr.log.critical(message)
                    raise ValueError(message)

                if outvalue is not None:
                    contents.append((url.normurl(url.join(invalue_dir, datafile)),
                                     url.normurl(url.join(outvalue_dir, datafile))))
                else:
                    contents.append(url.normurl(url.join(invalue_dir, datafile)))
        except Exception:
            message = 'Cannot determine ElementDataFile for {}'.format(value)
            fastr.log.critical(message)
            raise ValueError(message)

        return contents



import os
import imp

import nose.tools as nt

import fastr


def test_examples():
    examples_dir = fastr.config.examplesdir
    examples = [os.path.join(examples_dir, x) for x in os.listdir(examples_dir) if x.endswith(".py")]
    for example_path in examples:
        if not example_path.endswith("__init__.py"):
            yield run, example_path


def run(example_path):
    example_base, _ = os.path.splitext(os.path.basename(example_path))
    example_module = imp.load_source(example_base, example_path)
    network = example_module.create_network()
    source_data = example_module.source_data(network)
    sink_data = example_module.sink_data(network)
    result = network.execute(source_data, sink_data)
    nt.ok_(result)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastr


def create_network():
    network = fastr.Network(id_="test_cv")

    source1 = network.create_source('Int', id_='numbers')

    cv_node = network.create_node('CrossValidation')
    cv_node.inputs['items'] = source1.output
    cv_node.inputs['method'] = ['KFold']
    cv_node.inputs['number_of_folds'] = [3]
    fastr.log.info('[TEST] train output type: {}'.format(type(cv_node.outputs['train']).__name__))
    fastr.log.info('[TEST] test output type: {}'.format(type(cv_node.outputs['test']).__name__))
    cv_node.outputs['train'].preferred_types = fastr.datatypes.Int
    cv_node.outputs['test'].preferred_types = fastr.datatypes.Int

    product_node = network.create_node('Multiply')
    product_node.inputs['left_hand'] = cv_node.outputs['train']
    product_node.inputs['right_hand'] = cv_node.outputs['test']
    product_node.inputs['right_hand'].input_group = 'right'
    product_node.outputs['result'].preferred_types = fastr.datatypes.Float
    product_node.merge_dimensions = 'all'

    sum_node = network.create_node('Sum')
    link = network.create_link(product_node.outputs['result'], sum_node.inputs['values'])
    link.collapse = 'numbers_train'

    sink = network.create_sink('Int', id_='sink')
    sink.input = sum_node.outputs['result']

    return network


def source_data(network):
    return {'numbers': list(range(5))}


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    print(""" WARNING: experimental feature ;) """)
    print(""" Please run "pip install sklearn" to make use of the experimental cross validation flow. """)
    network = create_network()

    network.draw_network(img_format='svg', draw_dimension=True)

    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains function to help gather system information use for the
provenance of the Job execution.
"""

import getpass
import os
import platform
import sys

try:
    import psutil
    _PSUTIL_AVAILABLE = True
except ImportError:
    _PSUTIL_AVAILABLE = False


def namedtuple_to_dict(ntuple):
    """
    Helper function to convert a named tuple into a dict

    :param namedtuple ntuple: the namedtuple to convert
    :return: named tuple as a dict
    :rtype: dict
    """
    # Pylint does not play nice with namedtuples (does not recognize their methods)
    # pylint: disable=no-member,maybe-no-member,protected-access
    return dict(ntuple._asdict())


def get_cpu_usage():
    """
    Get the current CPU usage

    :return: CPU usage info
    :rtype: dict
    """
    cpu_usage = namedtuple_to_dict(psutil.cpu_times())
    cpu_usage['percent'] = psutil.cpu_percent()

    return cpu_usage


def get_memory_usage():
    """
    Get the current memory usage

    :return: memory usage info
    :rtype: dict
    """
    virtual_memory = namedtuple_to_dict(psutil.virtual_memory())
    swap_memory = namedtuple_to_dict(psutil.swap_memory())

    return {
            'virtual': virtual_memory,
            'swap': swap_memory}


def get_mounts():
    """
    Get the current mounts known on the system

    :return: mount info
    :rtype: dict
    """
    mounts = []
    for mount in psutil.disk_partitions():
        mount = namedtuple_to_dict(mount)

        try:
            usage = psutil.disk_usage(mount['mountpoint'])
            mount["disk_total"] = usage.total
            mount["disk_used"] = usage.used
            mount["disk_free"] = usage.free
            mount["disk_use_percent"] = usage.percent
        except OSError:
            pass

        mounts.append(mount)

    return mounts


def get_users():
    """
    Get current users on the system

    :return: user info
    :rtype: dict
    """
    users = []
    for user in psutil.users():
        user = namedtuple_to_dict(user)
        users.append(user)

    return users


def get_processes():
    """
    Get a list of all currently running processes

    :return: process information
    :rtype: list
    """
    # List all running processes
    processes = []
    for process in psutil.process_iter():
        proc = {'pid': process.pid,
                'name': process.name,
                'status': process.status,
                'memoryuse': process.get_memory_percent(),
                'cpu_user': process.get_cpu_times()[0],
                'cpu_system': process.get_cpu_times()[1],
                'nrthreads': process.get_num_threads()}
        try:
            proc["user"] = process.username
        except psutil.AccessDenied:
            pass
        try:
            proc["workdir"] = process.getcwd()
        except psutil.AccessDenied:
            pass
        try:
            proc["nice"] = process.get_nice()
        except psutil.AccessDenied:
            pass

        processes.append(proc)

    return processes


def get_sysinfo():
    """
    Get system information (cpu, memory, mounts and users)

    :return: system information
    :rtype: dict
    """
    if _PSUTIL_AVAILABLE:
        sysinfo = {'cpu_usage': get_cpu_usage(),
                   'memory_usage': get_memory_usage(),
                   'mounts': get_mounts(),
                   'users': get_users()}
    else:
        sysinfo = {}
    return sysinfo


def get_os():
    """
    Get information about the OS

    :return: OS information
    :rtype: dict
    """
    os_ = {'system': platform.system(),
           'is64bits': int(sys.maxsize > 2 ** 32),
           'maxsize': sys.maxsize}

    version = {}

    # Save version depending on platform
    if platform.system() == 'Linux':
        linux_distribution = platform.linux_distribution()
        version["distname"] = linux_distribution[0]
        version["version"] = linux_distribution[1]
        version["id"] = linux_distribution[2]

        libc = dict(zip(['libname', 'version'], platform.libc_ver()))
        version["libc"] = libc
    elif platform.system() == 'Windows':
        win32_version = platform.win32_ver()
        version["release"] = win32_version[0]
        version["version"] = win32_version[1]
        version["servicepack"] = win32_version[2]
        version["ptype"] = win32_version[3]
    elif platform.system() == 'Java':
        java_ver = platform.java_ver()
        version["release"] = java_ver[0]
        version["vendor"] = java_ver[1]
        version["vm_name"] = java_ver[2][0]
        version["vm_release"] = java_ver[2][1]
        version["vm_vendor"] = java_ver[2][2]
        version["os_name"] = java_ver[3][0]
        version["os_version"] = java_ver[3][1]
        version["os_arch"] = java_ver[3][2]
    elif platform.system() == 'Darwin':
        mac_ver = platform.mac_ver()
        version["release"] = mac_ver[0]
        version["version"] = mac_ver[1][0]
        version["dev_stage"] = mac_ver[1][1]
        version["non_release_version"] = mac_ver[1][2]
        version["machine"] = mac_ver[2]

    os_['version'] = version

    return os_


def get_python():
    """
    Get information about the currently used Python implementation

    :return: python info
    :rtype: dict
    """
    return {"version": platform.python_version(),
            "branch": platform.python_branch(),
            "compiler": platform.python_compiler(),
            "implementation": platform.python_implementation(),
            "revision": platform.python_revision()}


def get_drmaa_info():
    """
    Get information about the SGE cluster (if applicable)

    :return: cluster info
    :rtype: dict
    """
    drmaa_info = {}
    drmaa_info['jobid'] = os.environ.get('JOB_ID', None)
    drmaa_info['taskid'] = os.environ.get('SGE_TASK_ID', None)
    drmaa_info['jobname'] = os.environ.get('JOB_NAME', None)
    return drmaa_info


def get_hostinfo():
    """
    Get all information about the current host machine

    :return: host info
    :rtype: dict
    """
    hostenv = {}
    hostenv['os'] = get_os()
    hostenv['uname'] = platform.uname()
    hostenv['username'] = getpass.getuser()
    hostenv['cwd'] = os.getcwd()
    hostenv['python'] = get_python()
    hostenv['envvar'] = os.environ
    hostenv['drmaa'] = get_drmaa_info()
    return hostenv



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nose.tools as nt

import fastr
from fastr.core.node import Node, SourceNode, SinkNode, ConstantNode
from fastr.utils.dicteq import diffobj_str


class TestNode():

    def setup(self):
        self.source_network = fastr.Network('source_network')
        self.addint = fastr.Node(fastr.toollist['AddInt'], 'testnode_addint')
        self.sum = fastr.Node(fastr.toollist['Sum'], 'testnode_addimages')
        self.source_node = fastr.SourceNode('Int', 'source')
        self.source_node2 = fastr.SourceNode('Int', 'source2')
        self.sink_node = fastr.SinkNode('Int', 'sink')
        self.sink_node2 = fastr.SinkNode('Int', 'sink2')
        self.constant_node = fastr.ConstantNode('Int', [42], 'constant')
        self.constant_node2 = fastr.ConstantNode('Int', [42], 'constant2')

        fastr.log.info('Target network')
        self.target_network = fastr.Network('target_network')
        fastr.log.info('Created')

    def test_node_eq_operator(self):
        nt.eq_(self.addint, self.addint)
        nt.eq_(self.sum, self.sum)
        nt.ok_(self.addint != self.sum)
        nt.ok_(self.sum != self.addint)

    def test_sourcenode_eq_operator(self):
        nt.eq_(self.source_node, self.source_node)
        nt.eq_(self.source_node2, self.source_node2)
        nt.ok_(self.source_node != self.source_node2)
        nt.ok_(self.source_node2 != self.source_node)

    def test_constantnode_eq_operator(self):
        nt.eq_(self.constant_node, self.constant_node)
        nt.eq_(self.constant_node2, self.constant_node2)
        nt.ok_(self.constant_node != self.constant_node2)
        nt.ok_(self.constant_node2 != self.constant_node)

    def test_sinknode_eq_operator(self):
        nt.eq_(self.sink_node, self.sink_node)
        nt.eq_(self.sink_node2, self.sink_node2)
        nt.ok_(self.sink_node != self.sink_node2)
        nt.ok_(self.sink_node2 != self.sink_node)

    def test_node_getstate_setstate(self):
        #  Dump state and rebuild in a clean object
        addint_state = self.addint.__getstate__()
        addint_rebuilt = Node.__new__(Node)
        addint_rebuilt.__setstate__(addint_state)
        nt.eq_(self.addint, addint_rebuilt, msg=diffobj_str(self.addint, addint_rebuilt))

        #  Dump state and rebuild in a clean object
        addimages_state = self.sum.__getstate__()
        addimages_rebuilt = Node.__new__(Node)
        addimages_rebuilt.__setstate__(addimages_state)
        nt.eq_(self.sum, addimages_rebuilt, msg=diffobj_str(self.sum, addimages_rebuilt))

    def test_node_serializing(self):
        addint_json = self.addint.dumps()
        addimages_json = self.sum.dumps()
        addint_rebuilt = Node.loads(addint_json)
        addimages_rebuilt = Node.loads(addimages_json)
        nt.eq_(self.addint, addint_rebuilt, msg=diffobj_str(self.addint, addint_rebuilt))
        nt.eq_(self.sum, addimages_rebuilt, msg=diffobj_str(self.sum, addimages_rebuilt))
        nt.ok_(self.sum != addint_rebuilt)
        nt.ok_(self.addint != addimages_rebuilt)
        nt.ok_(addint_rebuilt != addimages_rebuilt)

    def test_source_node_serializing(self):
        source_json = self.source_node.dumps()
        fastr.log.info('JSON: {}'.format(source_json))
        source_rebuilt = SourceNode.loads(source_json)
        nt.eq_(self.source_node, source_rebuilt, msg=diffobj_str(self.source_node, source_rebuilt))

    def test_sink_node_serializing(self):
        sink_json = self.sink_node.dumps()
        fastr.log.info('JSON: {}'.format(sink_json))
        sink_rebuilt = SinkNode.loads(sink_json)
        nt.eq_(self.sink_node, sink_rebuilt, msg=diffobj_str(self.sink_node, sink_rebuilt))

    def test_constant_node_serializing(self):
        constant_json = self.constant_node.dumps()
        fastr.log.info('JSON: {}'.format(constant_json))
        constant_rebuilt = ConstantNode.loads(constant_json)
        nt.eq_(self.constant_node, constant_rebuilt, msg=diffobj_str(self.constant_node, constant_rebuilt))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import Queue
import sys
import threading

import fastr
from fastr.core.baseplugin import PluginState

try:
    import drmaa
    load_drmaa = True
except (ImportError, RuntimeError):
    load_drmaa = False

from fastr.execution.executionpluginmanager import ExecutionPlugin
from fastr.utils.classproperty import classproperty


class DRMAAExecution(ExecutionPlugin):
    """
    A DRMAA execution plugin to execute Jobs on a Grid Engine cluster. It uses
    a configuration option for selecting the queue to submit to. It uses the
    python ``drmaa`` package.

    .. note::

        To use this plugin, make sure the ``drmaa`` package is installed and
        that the execution is started on an SGE submit host with DRMAA
        libraries installed.

    .. note::

        This plugin is at the moment tailored to SGE, but it should be fairly
        easy to make different subclasses for different DRMAA supporting
        systems.
    """
    if load_drmaa:
        try:
            # Try to create a session
            session = drmaa.Session()
            session.initialize()
            session.exit()
            del session
        except:
            _status = (PluginState.failed, 'Could not create a DRMAA session')
    else:
        _status = (PluginState.failed, 'Could not load DRMAA module required for cluster communication')

    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        super(DRMAAExecution, self).__init__(finished_callback, cancelled_callback, status_callback)

        # Some default
        self.default_queue = fastr.config.drmaa_queue

        # Create the DRMAA session
        self.session = drmaa.Session()
        self.session.initialize()
        fastr.log.info('A DRMAA session was started successfully')
        response = self.session.contact
        fastr.log.info('session contact returns: ' + response)

        # Create job translation table
        self.job_translation_table = dict()
        self.job_lookup_table = dict()

        # Create even queue lock
        self.submit_queue = Queue.Queue()
        self.queue_lock = threading.Event()

        # Create callback collector and job submitter
        self.running = True
        fastr.log.debug('Creating job collector')
        self.collector = threading.Thread(name='DRMAAJobCollector-0', target=self.collect_jobs, args=())
        self.collector.daemon = True
        fastr.log.debug('Starting job collector')
        self.collector.start()

        fastr.log.debug('Creating job submitter')
        self.submitter = threading.Thread(name='DRMAAJobSubmitter-0', target=self.submit_jobs, args=())
        self.submitter.daemon = True
        fastr.log.debug('Starting job submitter')
        self.submitter.start()

    @classproperty
    def configuration_fields(cls):
        return {
            "drmaa_queue": (str, "week", "The default queue to use for jobs send to the scheduler")
        }

    def cleanup(self):
        # Stop submissions and callbacks
        self.running = False  # Signal collector thread to stop running
        super(DRMAAExecution, self).cleanup()

        # See if there are leftovers in the job translation table that can be cancelled
        while len(self.job_translation_table) > 0:
            drmaa_job_id, job = self.job_translation_table.popitem()
            fastr.log.info('Terminating left-over job {}'.format(drmaa_job_id))
            self.session.control(drmaa_job_id, 'terminate')

        fastr.log.debug('Stopping DRMAA executor')
        # Destroy DRMAA
        try:
            self.session.exit()
            fastr.log.debug('Exiting DRMAA session')
        except drmaa.NoActiveSessionException:
            pass
        if self.collector.isAlive():
            fastr.log.debug('Terminating job collector thread')
            self.collector.join()
        if self.submitter.isAlive():
            fastr.log.debug('Terminating job submitter thread')
            self.submitter.join()
        fastr.log.debug('DRMAA executor stopped!')

    def _queue_job(self, job):
        self.submit_queue.put(job, block=True)

    def _cancel_job(self, job):
        if job.id not in self.job_lookup_table:
            fastr.log.info('Job {} not found in DRMAA lookup'.format(job.id))
            return

        drmaa_job_id = self.job_lookup_table[job.id]
        fastr.log.info('Cancelling job {}'.format(drmaa_job_id))
        try:
            self.session.control(drmaa_job_id, 'terminate')
        except drmaa.InvalidJobException:
            fastr.log.warning('Trying to cancel an unknown job, already finished/cancelled?')

        del self.job_translation_table[drmaa_job_id]

    def _release_job(self, job):
        pass

    def _job_finished(self, result):
        pass

    # FIXME This needs to be more generic! This is for our SGE cluster only!
    def send_job(self, command, arguments, queue=None, walltime=None, job_name=None, memory=None, ncores=None, joinLogFiles=False, outputLog=None, errorLog=None, hold_job=None):

        # Create job template
        jt = self.session.createJobTemplate()
        jt.remoteCommand = command

        jt.args = arguments
        jt.joinFiles = joinLogFiles
        env = os.environ
        # Make sure environment modules do not annoy use with bash warnings
        # after the shellshock bug was fixed
        env.pop('BASH_FUNC_module()', None)
        jt.jobEnvironment = env

        if queue is None:
            queue = self.default_queue

        native_spec = '-cwd -q %s' % queue

        if walltime is not None:
            native_spec += ' -l h_rt=%s' % walltime

        if memory is not None:
            native_spec += ' -l h_vmem=%s' % memory

        if ncores is not None:
            native_spec += ' -pe smp %d' % ncores

        if outputLog is not None:
            native_spec += ' -o %s' % outputLog

        if errorLog is not None:
            native_spec += ' -e %s' % errorLog

        if hold_job is not None:
            if isinstance(hold_job, int):
                native_spec += ' -hold_jid {}'.format(hold_job)
            elif isinstance(hold_job, list) or isinstance(hold_job, tuple):
                if len(hold_job) > 0:
                    jid_list = ','.join([str(x) for x in hold_job])
                    native_spec += ' -hold_jid {}'.format(jid_list)
            else:
                fastr.log.error('Incorrect hold_job type!')

        fastr.log.debug('Setting native spec to: {}'.format(native_spec))
        jt.nativeSpecification = native_spec
        if job_name is None:
            job_name = command
            job_name = job_name.replace(' ', '_')
            job_name = job_name.replace('"', '')
            if len(job_name) > 32:
                job_name = job_name[0:32]

        jt.jobName = job_name

        # Send job to cluster
        jobid = self.session.runJob(jt)

        # Remove job template
        self.session.deleteJobTemplate(jt)

        return jobid

    def submit_jobs(self):
        while self.running:
            try:
                job = self.submit_queue.get(block=True, timeout=2)

                # Get job command and write to file
                command = [sys.executable,
                           os.path.join(fastr.config.executionscript),
                           job.commandfile]
                fastr.log.debug('Command to queue: {}'.format(command))

                # Make sure we do not submit after it stopped running
                if not self.running:
                    break

                fastr.log.info('Queueing {} via DRMAA'.format(job.jobid))

                # Submit command to scheduler
                cl_job_id = self.send_job(command[0], command[1:], job_name='fastr_{}'.format(job.jobid),
                                          memory=job.required_memory,
                                          ncores=job.required_cores,
                                          walltime=job.required_time,
                                          outputLog=job.stdoutfile,
                                          errorLog=job.stderrfile,
                                          hold_job=[self.job_lookup_table[x] for x in job.hold_jobs if x in self.job_lookup_table],
                                          )

                # Register job in the translation tables
                self.job_translation_table[cl_job_id] = job
                fastr.log.debug('Inserting {} in lookup table pointing to {}'.format(job.jobid, cl_job_id))
                self.job_lookup_table[job.jobid] = cl_job_id
                fastr.log.info('Job {} queued via DRMAA'.format(job.jobid))

                # Set the queue lock to indicate there is content in the queue
                if not self.queue_lock.is_set():
                    fastr.log.debug('Setting queue_lock')
                    self.queue_lock.set()
            except Queue.Empty:
                pass

        fastr.log.info('DRMAA submission thread ended!')

    def collect_jobs(self):
        first_wait = True
        second_wait = True

        while self.running:
            # Wait for the queue to contain
            if first_wait and not self.queue_lock.is_set():
                fastr.log.debug('Waiting jobs to be queued...')

            if not self.queue_lock.wait(2):
                first_wait = False
                continue

            first_wait = True

            if second_wait:
                fastr.log.debug('Waiting for a job to finish...')

            try:
                info = self.session.wait(drmaa.Session.JOB_IDS_SESSION_ANY, 5)
                second_wait = True
            except drmaa.ExitTimeoutException:
                second_wait = False
                continue
            except drmaa.InvalidJobException:
                fastr.log.debug('No valid jobs (session queue appears to be empty)')
                fastr.log.debug('Clearing queue_lock')
                self.queue_lock.clear()
                continue
            except drmaa.NoActiveSessionException:
                if not self.running:
                    fastr.log.info('DRMAA session no longer active, quiting collector...')
                else:
                    fastr.log.critical('DRMAA session no longer active, but DRMAA executor not stopped properly! Quitting')
                    self.running = False
                continue
            except drmaa.errors.DrmaaException as exception:
                # Avoid the collector getting completely killed on another DRMAA exception
                fastr.log.warning('Encountered unexpected DRMAA exception: {}'.format(exception))
                second_wait = True
                continue
            except Exception as exception:
                #if exception.message == 'code 24: no usage information was returned for the completed job':
                if exception.message.startswith('code 24:'):
                    # Avoid the collector getting completely killed this specific exception
                    fastr.log.warning('Encountered unexpected DRMAA exception: {}'.format(exception))
                    second_wait = True
                    continue
                else:
                    raise

            fastr.log.info('Cluster DRMAA job {} finished'.format(info.jobId))

            # Create a copy of the job that finished and remove from the translation table
            job = copy.deepcopy(self.job_translation_table.get(info.jobId))

            if job is not None:
                del self.job_translation_table[info.jobId]

                # Send the result to the callback function
                self.job_finished(job)
            else:
                fastr.log.warning('Job {} no longer available (got cancelled?)'.format(info.jobId))

        fastr.log.info('DRMAA collection thread ended!')



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import nose.tools as nt

import fastr
from fastr.core.tool import Tool
from fastr.utils.dicteq import diffdict
from fastr.core.version import Version
from fastr.utils.dicteq import diffobj_str


class TestTool():

    def setup(self):
        self.addint = fastr.toollist['AddInt']
        self.addimages = fastr.toollist['Sum']
        self.addint_copy = copy.deepcopy(self.addint)
        self.addimages_copy = copy.deepcopy(self.addimages)

    def test_tool_eq_operator(self):
        nt.eq_(self.addint, self.addint, msg=diffobj_str(self.addint, self.addint))
        nt.eq_(self.addint, self.addint_copy, msg=diffobj_str(self.addint, self.addint_copy))
        nt.eq_(self.addimages, self.addimages_copy, msg=diffobj_str(self.addimages, self.addimages_copy))
        assert self.addint != self.addimages
        assert self.addint_copy != self.addimages

    def test_tool_getstate_setstate(self):
        #  Dump state and rebuild in a clean object
        addint_state = self.addint.__getstate__()
        addint_rebuilt = Tool.__new__(Tool)
        addint_rebuilt.__setstate__(addint_state)
        nt.eq_(self.addint, addint_rebuilt)

        #  Dump state and rebuild in a clean object
        addimages_state = self.addimages.__getstate__()
        addimages_rebuilt = Tool.__new__(Tool)
        addimages_rebuilt.__setstate__(addimages_state)
        nt.eq_(self.addimages, addimages_rebuilt)

    def test_tool_serializing(self):
        addint_json = self.addint.dumps()
        addint_rebuilt = Tool.loads(addint_json)
        nt.eq_(self.addint, addint_rebuilt, msg='Differences:\n{}'.format('\n'.join(diffdict(self.addint.__dict__, addint_rebuilt.__dict__))))

    def test_toolversions(self):
        toolname = 'Source'
        tool = fastr.toollist[toolname]
        version = Version("1.0")
        nt.eq_(fastr.toollist.toolversions(tool)[0], version)
        nt.eq_(fastr.toollist.toolversions(toolname)[0], version)
        nt.eq_(len(fastr.toollist.toolversions(toolname)), 1, msg="Excpected number of versions for tool {} is 1 not {}.".format(toolname, len(fastr.toollist.toolversions(toolname))))
        # Maybe a unique id should be used here, but I think it very unlikely that this toolname will be encountered.
        nt.eq_(fastr.toollist.toolversions('097G)(A&*BD)(A7vA)(F67vAD)F967vAF-67vAD)F(86vAFD096vAF)97v'), None)
        assert isinstance(fastr.toollist.toolversions(tool), list), "%r is not of list type" % fastr.toollist.toolversions(tool)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module to compare various fastr specific things such as a execution directory
or a reference directory.
"""

import os

import fastr
from fastr import Network, SourceNode
from fastr.execution.job import Job
from fastr.execution.networkanalyzer import DefaultNetworkAnalyzer
from fastr.utils.iohelpers import load_gpickle


def compare_execution_dir(path1, path2):
    # Compare network dumps
    network_file1 = os.path.join(path1, Network.NETWORK_DUMP_FILE_NAME)
    network_file2 = os.path.join(path2, Network.NETWORK_DUMP_FILE_NAME)

    network1 = Network.loadf(network_file1)
    network2 = Network.loadf(network_file2)

    fastr.log.debug('Network1 filename: {}'.format(network1.filename))
    fastr.log.debug('Network2 filename: {}'.format(network2.filename))

    if network1 != network2:
        yield "Networks dumps are not equal!"
    else:
        del network2

    # Get the order of the Nodes
    analyzer = DefaultNetworkAnalyzer()
    execution_order = analyzer.analyze_network(
        network1,
        (None, network1.nodelist.values())
    )

    fastr.log.debug('Execution order: "{}"'.format(execution_order))

    # Compare node outputs in execution order
    for node in execution_order:
        fastr.log.debug('Checking node {}'.format(node.id))
        # Get the sample present
        node_dir1 = os.path.join(path1, node.id)
        node_dir2 = os.path.join(path2, node.id)

        if isinstance(node, SourceNode):
            # Possible source nodes do not exist
            if not os.path.isdir(node_dir1):
                if not os.path.isdir(node_dir2):
                    # Non-existing in both
                    continue
                else:
                    yield("Node '{}' does not have output for result 2")
                    continue
            elif not os.path.isdir(node_dir2):
                yield("Node '{}' does not have output for result 1")
                continue

        samples1 = set(os.listdir(node_dir1))
        samples2 = set(os.listdir(node_dir2))

        # Compare the samples
        if samples1 != samples2:
            yield ("Node '{}' contains different samples, set 1 exclusively"
                   " contains {} and set 2 exclusively contains {}").format(
                node.id,
                samples1.difference(samples2),
                samples2.difference(samples1)
            )

        # Inspect the individual samples that are in both sets
        for sample in sorted(samples1.intersection(samples2)):
            fastr.log.debug('Checking sample {}'.format(sample))
            result1 = os.path.join(node_dir1, sample, Job.RESULT_DUMP)
            result2 = os.path.join(node_dir2, sample, Job.RESULT_DUMP)

            job1 = load_gpickle(result1)
            job2 = load_gpickle(result2)

            # Compare output data
            fastr.log.debug('Job1 data: {}'.format(job1.output_data))
            fastr.log.debug('Job2 data: {}'.format(job2.output_data))
            if job1.output_data != job2.output_data:
                yield "Output data for job {} does not match".format(job1.jobid)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the object manager class
"""

from abc import abstractproperty, abstractmethod

import fastr
from fastr.core.basemanager import BaseManager
from fastr.core.version import Version
import fastr.exceptions as exceptions


class ObjectManager(BaseManager):
    """
    Class for managing all the objects loaded in the fastr system
    """
    def __init__(self, path):
        """
        Create a ObjectManager and scan path to search for Objects

        :param path: the path(s) to scan for Objects
        :type path: str or iterable of str
        :return: newly created ObjectManager
        """
        super(ObjectManager, self).__init__(path, True)

    def __contains__(self, key):
        """
        Check if an item is in the ObjectManager

        :param key: object id or tuple (Objectid, version)
        :type key: str or tuple
        :return: flag indicating the item is in the manager
        """
        return self.__keytransform__(key) in self.data

    def __getitem__(self, key):
        """
        Retrieve a Object from the ObjectManager. You can request by only an id,
        which results in the newest version of the Object being returned, or
        request using both an id and a version.

        :param key: object id or tuple (Objectid, version)
        :type key: str or tuple
        :return: the requested Object
        :raises FastrObjectUnknownError: if a non-existing Object was requested
        """
        new_key = self.__keytransform__(key)
        if new_key not in self.data:
            raise exceptions.FastrObjectUnknownError('Key "{}" (expanded to {}) not found in {}'.format(key, new_key, type(self).__name__))

        obj = self.data[new_key]
        return obj

    def __keytransform__(self, key):
        """
        Key transform, used for allowing indexing both by id-only and by
        ``(id, version)``

        :param key: key to transform
        :return: key in form ``(id, version)``
        """
        # Get the version (or None)
        if isinstance(key, (str, unicode)):
            namespace, object_id, version = None, key, None
        elif isinstance(key, tuple):
            if len(key) == 2:
                object_id, version = key
                namespace = None
            elif len(key) == 3:
                namespace, object_id, version = key
            else:
                raise exceptions.FastrValueError('Object key in tuple for should be length 2 or 3, '
                                                 'in the form (id, version) or (namespace, id, version)')
        else:
            raise exceptions.FastrTypeError('Object key should be a string, unicode or tuple!')

        # Split out version first (it may contain./, whereas namespace and object won't contain /)
        if version is None and '/' in object_id:
            object_id, version = object_id.split('/', 1)

        # Split out namespace
        if namespace is None and '.' in object_id:
            namespace, object_id = object_id.rsplit('.', 1)

        if isinstance(version, (str, unicode)):
            version = Version(version)

        # Check that we are done (should be unique match)
        if all(x is not None for x in [namespace, object_id, version]):
            return namespace, object_id, version

        # Check the namespace, if there is only 1 namespace with a object, use that
        if namespace is None:
            namespaces = {k[0] for k in self.keys() if k[1] == object_id}

            if len(namespaces) == 0:
                return namespace, object_id, version
            elif len(namespaces) > 1:
                raise exceptions.FastrValueError('Multiple versions of Object {} found in different namespaces,'.format(object_id) +
                                                 ' need to pick an explicit namespace! (namespace: {})'.format(sorted(namespaces)))

            namespace = namespaces.pop()

        # If the version is not given, figure it out using newest version available
        if version is None:
            versions = [k[2] for k in self.keys() if k[0] == namespace and k[1] == object_id]

            if len(versions) > 0:
                version = max(versions)

        return namespace, object_id, version

    @property
    def _item_extension(self):
        """
        Extension of files to load
        """
        return '(xml|json)'

    @abstractproperty
    def object_class(self):
        """
        The class of the objects to populate the manager with
        """

    @abstractmethod
    def get_object_version(self, obj):
        """
        Get the version of a given object

        :param object: the object to use
        :return: the version of the object
        """

    def _print_key(self, key):
        """
        Print function for the keys
        """
        object_id = key[0] if key[0] else ''
        if object_id != '':
            object_id += '.'
        object_id += key[1]
        return (object_id, 'v{}'.format(str(key[2])))

    def _print_value(self, value):
        """
        Print function for the values
        """
        return value.filename

    def _load_item(self, filepath, namespace):
        """
        Load a Object file and store it in the Manager
        """
        obj = self.object_class.loadf(filepath)
        object_version = self.get_object_version(obj)
        if namespace.endswith('{}'.format(object_version)):
            namespace = namespace[:-len(str(object_version))]
            namespace = namespace.strip('.')
        self._store_item((namespace, obj.id, object_version), obj)
        obj.namespace = namespace

    def todict(self):
        """
        Return a dictionary version of the Manager

        :return: manager as a dict
        """
        result = {}
        for key in self.keys():
            if key[0] not in result:
                result[key[0]] = []

            if str(key[1]) not in result[key[0]]:
                result[key[0]].append(str(key[1]))

        return result

    def objectversions(self, obj):
        """
        Return a list of available versions for the object

        :param object: The object to check the versions for. Can be either a `Object` or a `str`.
        :return: List of version objects. Returns `None` when the given object is not known.
        """

        if isinstance(obj, self.object_class):
            objectname = obj.id
            namespace = obj.namespace
        else:
            try:
                obj = self[obj]
                objectname = obj.id
                namespace = obj.namespace
            except exceptions.FastrObjectUnknownError as e:
                fastr.log.error("Error requesting object versions: Object ({}) is not known.".format(obj))
                return None
        fastr.log.info("objectname: {}".format(objectname))
        return sorted(key[2] for key in self.keys() if key[0] == namespace and key[1] == objectname)



from prov.model import ProvDocument
import fastr


class Provenance(object):
    """
    The Provenance object keeps track of everything that happens to a data object.
    """

    def __init__(self, parent, host=None):
        self.parent = parent
        if host is not None:
            self.host = host.rstrip('/')
        else:
            self.host = fastr.config.web_url()
        self.namespaces = {}
        self.document = ProvDocument()

        # Define default namespaces
        self.fastr = self._add_namespace('fastr')
        self.tool = self._add_namespace('tool')
        self.node = self._add_namespace('node')
        self.job = self._add_namespace('job')
        self.data = self._add_namespace('data')
        self.worker = self._add_namespace('worker')
        self.network = self._add_namespace('network')
        self.fastr_info = self._add_namespace('fastrinfo')

        # Add fastr agent
        self.fastr_agent = self.agent(self.fastr[fastr.version.version])

    def _add_namespace(self, name, parent=None, url=None):
        if parent is None:
            host = self.host
        else:
            if parent in self.namespaces.keys():
                host = self.namespaces[parent].uri
            else:
                return None

        if url is None:
            self.namespaces[name] = self.document.add_namespace(name, "{}/{}/".format(host, name))
        else:
            self.namespaces[name] = self.document.add_namespace(name, "{}/{}".format(host, url))
        return self.namespaces[name]

    def agent(self, identifier, other_attributes=None):
        return self.document.agent(identifier, other_attributes)

    def activity(self, identifier, start_time=None, end_time=None, other_attributes=None):
        return self.document.activity(identifier, start_time, end_time, other_attributes)

    def entity(self, identifier, other_attributes=None):
        return self.document.entity(identifier, other_attributes)

